"""Post-processing - link fixing and redirect generation."""

from __future__ import annotations

import logging

from sqlalchemy import select

from ...models import Entity
from .link_fixer import InternalLinkFixer, URLMapBuilder
from .redirects import RedirectGenerator

logger = logging.getLogger(__name__)


class PostProcessMixin:
    """Post-import processing methods: link fixing and redirect generation.

    Mixed into WordPressImportService.
    Requires: self.db, self.entity_svc, self.update_job(),
              self._get_entity_field()
    """

    async def fix_links(
        self,
        job_id: str,
        source_domain: str | None = None,
    ) -> dict:
        """
        Fix internal links in imported content.

        Rewrites WordPress URLs to Focomy URLs in post/page content.

        Args:
            job_id: Import job ID
            source_domain: Original WordPress domain for URL matching

        Returns:
            Dict with fix counts and details
        """
        from ...models import EntityValue

        await self.update_job(
            job_id,
            progress_message="Building URL map...",
        )

        # Build URL map
        builder = URLMapBuilder(self.db)
        url_map = await builder.build_map(source_domain)

        if not url_map:
            logger.info("No URL mappings found, skipping link fix")
            return {
                "success": True,
                "posts_fixed": 0,
                "pages_fixed": 0,
                "total_links_fixed": 0,
            }

        await self.update_job(
            job_id,
            progress_message=f"Fixing links with {len(url_map)} URL mappings...",
        )

        # Create link fixer
        fixer = InternalLinkFixer(url_map, source_domain)

        posts_fixed = 0
        pages_fixed = 0
        total_links_fixed = 0

        # Fix posts
        posts_result = await self.db.execute(
            select(Entity)
            .join(EntityValue, Entity.id == EntityValue.entity_id)
            .where(
                Entity.type == "post",
                EntityValue.field_name == "wp_id",
                Entity.deleted_at.is_(None),
            )
        )
        posts = posts_result.scalars().unique().all()

        for post in posts:
            content_result = await self.db.execute(
                select(EntityValue).where(
                    EntityValue.entity_id == post.id,
                    EntityValue.field_name == "content",
                )
            )
            content_ev = content_result.scalar_one_or_none()

            if content_ev and content_ev.value_text:
                fix_result = fixer.fix_content(content_ev.value_text)
                if fix_result.had_fixes:
                    content_ev.value_text = fix_result.content
                    posts_fixed += 1
                    total_links_fixed += len(fix_result.fixes)

        # Fix pages
        pages_result = await self.db.execute(
            select(Entity)
            .join(EntityValue, Entity.id == EntityValue.entity_id)
            .where(
                Entity.type == "page",
                EntityValue.field_name == "wp_id",
                Entity.deleted_at.is_(None),
            )
        )
        pages = pages_result.scalars().unique().all()

        for page in pages:
            content_result = await self.db.execute(
                select(EntityValue).where(
                    EntityValue.entity_id == page.id,
                    EntityValue.field_name == "content",
                )
            )
            content_ev = content_result.scalar_one_or_none()

            if content_ev and content_ev.value_text:
                fix_result = fixer.fix_content(content_ev.value_text)
                if fix_result.had_fixes:
                    content_ev.value_text = fix_result.content
                    pages_fixed += 1
                    total_links_fixed += len(fix_result.fixes)

        await self.db.commit()

        logger.info(
            f"Fixed links: {posts_fixed} posts, {pages_fixed} pages, "
            f"{total_links_fixed} total links"
        )

        await self.update_job(
            job_id,
            progress_message=f"Fixed {total_links_fixed} internal links",
        )

        return {
            "success": True,
            "posts_fixed": posts_fixed,
            "pages_fixed": pages_fixed,
            "total_links_fixed": total_links_fixed,
        }

    async def generate_redirects(
        self,
        job_id: str,
        source_url: str,
        config: dict | None = None,
    ) -> dict:
        """
        Generate URL redirects for imported content.

        Args:
            job_id: Import job ID
            source_url: Original WordPress site URL
            config: Optional configuration overrides

        Returns:
            Dict with redirect report data
        """
        from ...models import EntityValue

        await self.update_job(
            job_id,
            progress_message="Generating redirects...",
        )

        config = config or {}

        # Collect data from imported entities
        posts_data = []
        categories_data = []
        tags_data = []
        authors_data = []

        # Get posts
        posts_result = await self.db.execute(
            select(Entity)
            .join(EntityValue, Entity.id == EntityValue.entity_id)
            .where(
                Entity.type == "post",
                EntityValue.field_name == "wp_id",
                Entity.deleted_at.is_(None),
            )
        )
        posts = posts_result.scalars().unique().all()

        for post in posts:
            slug = await self._get_entity_field(post.id, "slug")
            title = await self._get_entity_field(post.id, "title")
            if slug:
                posts_data.append({
                    "old_url": f"/{slug}",
                    "new_slug": slug,
                    "slug": slug,
                    "title": title or slug,
                    "post_type": "post",
                })

        # Get pages
        pages_result = await self.db.execute(
            select(Entity)
            .join(EntityValue, Entity.id == EntityValue.entity_id)
            .where(
                Entity.type == "page",
                EntityValue.field_name == "wp_id",
                Entity.deleted_at.is_(None),
            )
        )
        pages = pages_result.scalars().unique().all()

        for page in pages:
            slug = await self._get_entity_field(page.id, "slug")
            title = await self._get_entity_field(page.id, "title")
            if slug:
                posts_data.append({
                    "old_url": f"/{slug}",
                    "new_slug": slug,
                    "slug": slug,
                    "title": title or slug,
                    "post_type": "page",
                })

        # Get categories
        cats_result = await self.db.execute(
            select(Entity)
            .join(EntityValue, Entity.id == EntityValue.entity_id)
            .where(
                Entity.type == "category",
                EntityValue.field_name == "wp_id",
                Entity.deleted_at.is_(None),
            )
        )
        categories = cats_result.scalars().unique().all()

        for cat in categories:
            slug = await self._get_entity_field(cat.id, "slug")
            name = await self._get_entity_field(cat.id, "name")
            if slug:
                categories_data.append({
                    "slug": slug,
                    "old_slug": slug,
                    "new_slug": slug,
                    "name": name or slug,
                })

        # Get tags
        tags_result = await self.db.execute(
            select(Entity)
            .join(EntityValue, Entity.id == EntityValue.entity_id)
            .where(
                Entity.type == "tag",
                EntityValue.field_name == "wp_id",
                Entity.deleted_at.is_(None),
            )
        )
        tags = tags_result.scalars().unique().all()

        for tag in tags:
            slug = await self._get_entity_field(tag.id, "slug")
            name = await self._get_entity_field(tag.id, "name")
            if slug:
                tags_data.append({
                    "slug": slug,
                    "old_slug": slug,
                    "new_slug": slug,
                    "name": name or slug,
                })

        # Get authors
        authors_result = await self.db.execute(
            select(Entity)
            .join(EntityValue, Entity.id == EntityValue.entity_id)
            .where(
                Entity.type == "user",
                EntityValue.field_name == "wp_id",
                Entity.deleted_at.is_(None),
            )
        )
        authors = authors_result.scalars().unique().all()

        for author in authors:
            login = await self._get_entity_field(author.id, "login")
            display_name = await self._get_entity_field(author.id, "display_name")
            if login:
                authors_data.append({
                    "login": login,
                    "new_slug": login,
                    "display_name": display_name or login,
                })

        # Generate redirects
        generator = RedirectGenerator(
            old_base_url=source_url,
            new_base_url=config.get("new_base_url", ""),
        )

        report = generator.generate_all(
            posts=posts_data,
            categories=categories_data,
            tags=tags_data,
            authors=authors_data,
            config=config,
        )

        logger.info(f"Generated {len(report.redirects)} redirects")

        await self.update_job(
            job_id,
            progress_message=f"Generated {len(report.redirects)} redirects",
        )

        return {
            "success": True,
            "redirect_count": len(report.redirects),
            "conflict_count": len(report.conflicts),
            "warnings": report.warnings,
            "redirects": [
                {
                    "from": r.from_path,
                    "to": r.to_path,
                    "status": r.status_code,
                    "regex": r.regex,
                    "comment": r.comment,
                }
                for r in report.redirects
            ],
            "conflicts": report.conflicts,
        }



"""Form API endpoints - public form handling."""

import json
from typing import Any

from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import HTMLResponse, RedirectResponse
from pydantic import BaseModel
from sqlalchemy.ext.asyncio import AsyncSession

from ..database import get_db
from ..engine.routes import render_theme
from ..rate_limit import limiter
from ..services.entity import EntityService
from ..services.mail import mail_service
from ..utils import require_feature_async

router = APIRouter(prefix="/forms", tags=["forms"])


class FormSubmitRequest(BaseModel):
    """Form submission request."""

    data: dict[str, Any]


@router.get("/{slug}", response_class=HTMLResponse)
async def view_form(
    slug: str,
    request: Request,
    db: AsyncSession = Depends(get_db),
):
    """View a public form."""
    await require_feature_async("form", db)
    entity_svc = EntityService(db)

    # Find form
    forms = await entity_svc.find(
        "form",
        limit=1,
        filters={"slug": slug, "status": "published"},
    )

    if not forms:
        raise HTTPException(status_code=404, detail="Form not found")

    form = forms[0]
    form_data = entity_svc.serialize(form)

    # Parse fields config
    fields_config = form_data.get("fields_config", [])
    if isinstance(fields_config, str):
        try:
            fields_config = json.loads(fields_config)
        except json.JSONDecodeError:
            fields_config = []

    # Parse steps config
    steps = form_data.get("steps", [])
    if isinstance(steps, str):
        try:
            steps = json.loads(steps)
        except json.JSONDecodeError:
            steps = []

    # Render form template (with admin bar context)
    html = await render_theme(
        db,
        "form.html",
        {
            "form": form_data,
            "fields": fields_config,
            "steps": steps,
            "csrf_token": getattr(request.state, "csrf_token", ""),
        },
        request=request,
        entity=form,
        content_type="form",
    )

    return HTMLResponse(content=html)


@router.post("/{slug}")
@limiter.limit("10/minute")
async def submit_form(
    slug: str,
    request: Request,
    db: AsyncSession = Depends(get_db),
):
    """Submit a form."""
    await require_feature_async("form", db)
    entity_svc = EntityService(db)

    # Find form
    forms = await entity_svc.find(
        "form",
        limit=1,
        filters={"slug": slug, "status": "published"},
    )

    if not forms:
        raise HTTPException(status_code=404, detail="Form not found")

    form = forms[0]
    form_data = entity_svc.serialize(form)

    # Get form data
    content_type = request.headers.get("content-type", "")

    if "application/json" in content_type:
        body = await request.json()
        submission_data = body.get("data", body)
    else:
        form_body = await request.form()
        submission_data = {k: v for k, v in form_body.items() if k != "csrf_token"}

    # Validate required fields
    fields_config = form_data.get("fields_config", [])
    if isinstance(fields_config, str):
        try:
            fields_config = json.loads(fields_config)
        except json.JSONDecodeError:
            fields_config = []

    errors = []
    for field in fields_config:
        if field.get("required") and not submission_data.get(field["name"]):
            errors.append(f"{field.get('label', field['name'])}は必須です")

    if errors:
        raise HTTPException(status_code=400, detail={"errors": errors})

    # Save submission
    await entity_svc.create(
        "form_submission",
        {
            "form_id": form.id,
            "form_title": form_data.get("title"),
            "data": submission_data,
            "ip_address": request.client.host if request.client else None,
            "user_agent": request.headers.get("User-Agent"),
            "status": "new",
        },
    )

    # Send notification email
    notify_email = form_data.get("notify_email")
    if notify_email:
        reply_to = submission_data.get("email")
        mail_service.send_form_notification(
            to=notify_email,
            form_title=form_data.get("title", ""),
            submission_data=submission_data,
            reply_to=reply_to,
        )

    # Handle response
    redirect_url = form_data.get("redirect_url")
    if redirect_url:
        return RedirectResponse(url=redirect_url, status_code=303)

    success_message = form_data.get("success_message", "送信が完了しました。")

    # Check if AJAX request
    if "application/json" in content_type:
        return {"success": True, "message": success_message}

    # Render success page (with admin bar context)
    html = await render_theme(
        db,
        "form_success.html",
        {
            "form": form_data,
            "message": success_message,
        },
        request=request,
    )

    return HTMLResponse(content=html)



"""Relation model - relationships as first-class citizens."""

from datetime import datetime
from typing import TYPE_CHECKING

from sqlalchemy import JSON, DateTime, ForeignKey, Index, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..database import Base

if TYPE_CHECKING:
    from .entity import Entity


class Relation(Base):
    """
    Relations between entities.

    No foreign keys hardcoded in entity tables.
    All relationships are managed through this single table.
    """

    __tablename__ = "relations"

    id: Mapped[int] = mapped_column(
        Integer,
        primary_key=True,
        autoincrement=True,
    )
    from_entity_id: Mapped[str] = mapped_column(
        String(36),
        ForeignKey("entities.id", ondelete="CASCADE"),
        nullable=False,
    )
    to_entity_id: Mapped[str] = mapped_column(
        String(36),
        ForeignKey("entities.id", ondelete="CASCADE"),
        nullable=False,
    )
    relation_type: Mapped[str] = mapped_column(
        String(100),
        nullable=False,
    )
    sort_order: Mapped[int] = mapped_column(
        Integer,
        default=0,
    )
    metadata_: Mapped[dict | None] = mapped_column(
        "metadata",
        JSON,
        nullable=True,
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime,
        nullable=False,
        default=datetime.utcnow,
    )

    # Relationships
    from_entity: Mapped["Entity"] = relationship(
        "Entity",
        foreign_keys=[from_entity_id],
        back_populates="relations_from",
    )
    to_entity: Mapped["Entity"] = relationship(
        "Entity",
        foreign_keys=[to_entity_id],
        back_populates="relations_to",
    )

    __table_args__ = (
        Index("idx_relations_from", "from_entity_id"),
        Index("idx_relations_to", "to_entity_id"),
        Index("idx_relations_type", "relation_type"),
        Index(
            "idx_relations_unique",
            "from_entity_id",
            "to_entity_id",
            "relation_type",
            unique=True,
        ),
    )

    def __repr__(self) -> str:
        return (
            f"<Relation({self.from_entity_id} -> {self.to_entity_id}, type={self.relation_type})>"
        )



"""Admin routes - Entity edit, update, and delete operations."""

import json as json_module
import logging

from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import HTMLResponse, RedirectResponse
from sqlalchemy.ext.asyncio import AsyncSession

from ..database import get_db
from ..models import Entity
from ..services.audit import AuditService, get_client_ip, get_request_id
from ..services.entity import EntityService
from ..services.field import field_service
from ..services.rbac import Permission
from .helpers import (
    check_permission,
    get_context,
    get_relation_options,
    require_admin,
    templates,
)
from .url import AdminURL

logger = logging.getLogger(__name__)

router = APIRouter()


@router.get("/{type_name}/{entity_id}/edit", response_class=HTMLResponse)
async def entity_edit(
    request: Request,
    type_name: str,
    entity_id: str,
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Edit entity form."""
    content_type = field_service.get_content_type(type_name)
    if not content_type:
        raise HTTPException(status_code=404, detail="Content type not found")

    entity_svc = EntityService(db)
    entity_raw = await entity_svc.get(entity_id)
    if not entity_raw or entity_raw.type != type_name:
        raise HTTPException(status_code=404, detail="Entity not found")

    entity = entity_svc.serialize(entity_raw)
    relations = await get_relation_options(type_name, entity_id, db)

    context = await get_context(request, db, current_user, type_name)
    context.update(
        {
            "type_name": type_name,
            "content_type": content_type.model_dump(),
            "entity": entity,
            "relations": relations,
            "message": request.query_params.get("message"),
            "form_action": AdminURL.entity_form_action(type_name, entity_id),
            "cancel_url": AdminURL.entity_list(type_name),
        }
    )

    return templates.TemplateResponse("admin/entity_form.html", context)


@router.post("/{type_name}/{entity_id}", response_class=HTMLResponse)
async def entity_update(
    request: Request,
    type_name: str,
    entity_id: str,
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Update entity."""
    content_type = field_service.get_content_type(type_name)
    if not content_type:
        raise HTTPException(status_code=404, detail="Content type not found")

    entity_svc = EntityService(db)
    entity_raw = await entity_svc.get(entity_id)
    if not entity_raw or entity_raw.type != type_name:
        raise HTTPException(status_code=404, detail="Entity not found")

    # RBAC permission check (includes ownership check for authors)
    await check_permission(db, current_user, type_name, Permission.UPDATE, entity_id)

    form_data = await request.form()

    # DEBUG: フォームデータをログ出力
    logger.info(f"[DEBUG] entity_update form_data keys: {list(form_data.keys())}")
    for k, v in form_data.items():
        if 'body' in k.lower() or 'content' in k.lower() or 'block' in k.lower():
            logger.info(f"[DEBUG] form_data[{k}] = {str(v)[:200] if v else 'None'}")

    # Build entity data from form
    data = {}
    for field in content_type.fields:
        value = form_data.get(field.name)
        if field.type == 'blocks':
            logger.info(f"[DEBUG] blocks field '{field.name}' value: {str(value)[:200] if value else 'None/Empty'}")
        if value is not None and value != "":
            # Type conversion
            if field.type in ("number", "integer"):
                data[field.name] = int(value)
            elif field.type == "float":
                data[field.name] = float(value)
            elif field.type == "boolean":
                data[field.name] = value == "true"
            elif field.type in ("blocks", "json", "multiselect"):
                try:
                    data[field.name] = json_module.loads(value)
                except (json_module.JSONDecodeError, TypeError):
                    data[field.name] = value
            elif field.type == "password":
                # Password is handled separately for user_auth table
                data["_password_plain"] = value
            elif field.type in ("media", "image"):
                # Handle file upload for media/image type
                if hasattr(value, 'file'):
                    from ..services.media import MediaService
                    media_svc = MediaService(db)
                    media = await media_svc.upload(
                        file=value.file,
                        filename=value.filename,
                        content_type=value.content_type,
                    )
                    data[field.name] = media_svc.serialize(media).get("url")
            else:
                data[field.name] = value
        elif field.type == "boolean":
            # Unchecked checkbox
            data[field.name] = False

    # Handle password update for user type
    password_plain = data.pop("_password_plain", None)
    if type_name == "user" and password_plain:
        from ..models.auth import UserAuth
        from sqlalchemy import select
        import bcrypt
        query = select(UserAuth).where(UserAuth.entity_id == entity_id)
        result = await db.execute(query)
        user_auth = result.scalar_one_or_none()
        if user_auth:
            salt = bcrypt.gensalt()
            user_auth.password_hash = bcrypt.hashpw(password_plain.encode(), salt).decode()

    # Save before state for audit
    before_data = entity_svc.serialize(entity_raw)

    try:
        user_data = entity_svc.serialize(current_user)
        await entity_svc.update(entity_id, data, user_id=user_data.get("id"))

        # Handle relations
        from ..services.relation import RelationService

        relation_svc = RelationService(db)

        for rel in content_type.relations:
            rel_field = f"rel_{rel.type}"
            rel_values = form_data.getlist(rel_field)
            # Filter out empty values
            rel_ids = [v for v in rel_values if v]
            # Sync relations (empty list = remove all)
            await relation_svc.sync(entity_id, rel_ids, rel.type)

        # Log update action
        entity_after = await entity_svc.get(entity_id)
        after_data = entity_svc.serialize(entity_after) if entity_after else data
        audit_svc = AuditService(db)
        await audit_svc.log_update(
            entity_type=type_name,
            entity_id=entity_id,
            entity_title=after_data.get("title") or after_data.get("name") or entity_id,
            before_data=before_data,
            after_data=after_data,
            user_id=user_data.get("id"),
            user_email=user_data.get("email"),
            user_name=user_data.get("name"),
            ip_address=get_client_ip(request),
            user_agent=request.headers.get("User-Agent"),
            request_id=get_request_id(request),
        )

        return RedirectResponse(
            url=f"/admin/{type_name}/{entity_id}/edit?message=Updated+successfully",
            status_code=303,
        )

    except ValueError as e:
        entity = entity_svc.serialize(entity_raw)
        entity.update(data)  # Show submitted values
        relations = await get_relation_options(type_name, entity_id, db)

        context = await get_context(request, db, current_user, type_name)
        context.update(
            {
                "type_name": type_name,
                "content_type": content_type.model_dump(),
                "entity": entity,
                "relations": relations,
                "error": str(e),
                "form_action": AdminURL.entity_form_action(type_name, entity_id),
                "cancel_url": AdminURL.entity_list(type_name),
            }
        )
        return templates.TemplateResponse("admin/entity_form.html", context)


@router.delete("/{type_name}/{entity_id}")
async def entity_delete(
    request: Request,
    type_name: str,
    entity_id: str,
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Delete entity (soft delete)."""
    entity_svc = EntityService(db)
    entity = await entity_svc.get(entity_id)
    if not entity or entity.type != type_name:
        raise HTTPException(status_code=404, detail="Entity not found")

    # RBAC permission check (includes ownership check for authors)
    await check_permission(db, current_user, type_name, Permission.DELETE, entity_id)

    # Save entity data for audit before deletion
    entity_data = entity_svc.serialize(entity)
    user_data = entity_svc.serialize(current_user)

    # Check if channel is protected from deletion
    if type_name == "channel":
        from ..services.channel import is_protected_channel

        if is_protected_channel(entity_data.get("slug", "")):
            raise HTTPException(
                status_code=400,
                detail="postsチャンネルは削除できません",
            )

    await entity_svc.delete(entity_id, user_id=user_data.get("id"))

    # Log delete action
    audit_svc = AuditService(db)
    await audit_svc.log_delete(
        entity_type=type_name,
        entity_id=entity_id,
        entity_title=entity_data.get("title") or entity_data.get("name") or entity_id,
        data=entity_data,
        user_id=user_data.get("id"),
        user_email=user_data.get("email"),
        user_name=user_data.get("name"),
        ip_address=get_client_ip(request),
        user_agent=request.headers.get("User-Agent"),
        request_id=get_request_id(request),
    )

    # Return empty response for HTMX to remove the row
    return HTMLResponse(content="", status_code=200)


@router.post("/{type_name}/{entity_id}/delete")
async def entity_delete_post(
    request: Request,
    type_name: str,
    entity_id: str,
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Delete entity via POST (for HTML forms)."""
    # Perform the delete
    entity_svc = EntityService(db)
    entity = await entity_svc.get(entity_id)
    if not entity or entity.type != type_name:
        raise HTTPException(status_code=404, detail="Entity not found")

    # RBAC permission check
    await check_permission(db, current_user, type_name, Permission.DELETE, entity_id)

    # Save entity data for audit before deletion
    entity_data = entity_svc.serialize(entity)
    user_data = entity_svc.serialize(current_user)

    # Check if channel is protected from deletion
    if type_name == "channel":
        from ..services.channel import is_protected_channel

        if is_protected_channel(entity_data.get("slug", "")):
            raise HTTPException(
                status_code=400,
                detail="postsチャンネルは削除できません",
            )

    await entity_svc.delete(entity_id, user_id=user_data.get("id"))

    # Log delete action
    audit_svc = AuditService(db)
    await audit_svc.log_delete(
        entity_type=type_name,
        entity_id=entity_id,
        entity_title=entity_data.get("title") or entity_data.get("name") or entity_id,
        data=entity_data,
        user_id=user_data.get("id"),
        user_email=user_data.get("email"),
        user_name=user_data.get("name"),
        ip_address=get_client_ip(request),
        user_agent=request.headers.get("User-Agent"),
        request_id=get_request_id(request),
    )

    # Redirect to list page for regular form submissions
    return RedirectResponse(
        url=f"/admin/{type_name}?message=Deleted+successfully",
        status_code=303,
    )



"""
Admin URL Helper - 管理画面URL一元管理

テンプレート内でのURL直書きを禁止し、
Python側でURL生成を行うことで一貫性を保つ。
"""


class AdminURL:
    """管理画面URL生成ヘルパー"""

    @staticmethod
    def entity_list(type_name: str, channel_slug: str | None = None) -> str:
        """エンティティ一覧URL"""
        if channel_slug:
            return f"/admin/channel/{channel_slug}/posts"
        return f"/admin/{type_name}"

    @staticmethod
    def entity_new(type_name: str, channel_slug: str | None = None) -> str:
        """エンティティ新規作成URL"""
        return f"{AdminURL.entity_list(type_name, channel_slug)}/new"

    @staticmethod
    def entity_edit(type_name: str, entity_id: int, channel_slug: str | None = None) -> str:
        """エンティティ編集URL"""
        base = AdminURL.entity_list(type_name, channel_slug)
        return f"{base}/{entity_id}/edit"

    @staticmethod
    def entity_form_action(
        type_name: str, entity_id: int | None = None, channel_slug: str | None = None
    ) -> str:
        """フォーム送信先URL（新規/更新共用）"""
        base = AdminURL.entity_list(type_name, channel_slug)
        if entity_id:
            return f"{base}/{entity_id}"
        return base

    @staticmethod
    def entity_delete(type_name: str, entity_id: int, channel_slug: str | None = None) -> str:
        """エンティティ削除URL"""
        base = AdminURL.entity_list(type_name, channel_slug)
        return f"{base}/{entity_id}"

    @staticmethod
    def entity_bulk(type_name: str, channel_slug: str | None = None) -> str:
        """一括操作URL - 常にtype_name形式（ルートが/{type_name}/bulkのため）"""
        # channel_slugがあっても、ルートは/{type_name}/bulk形式
        return f"/admin/{type_name}/bulk"

    @staticmethod
    def entity_pagination(
        type_name: str, page: int, channel_slug: str | None = None, query: str | None = None
    ) -> str:
        """ページネーションURL"""
        base = AdminURL.entity_list(type_name, channel_slug)
        url = f"{base}?page={page}"
        if query:
            url += f"&{query}"
        return url



"""Admin dashboard and media routes."""

from fastapi import APIRouter, Depends, Request
from fastapi.responses import HTMLResponse
from sqlalchemy.ext.asyncio import AsyncSession

from ..config import get_settings
from ..database import get_db
from ..models import Entity
from ..services.entity import EntityService
from ..services.field import field_service
from ..utils import require_feature
from .helpers import templates, require_admin, get_context

router = APIRouter()


# === Dashboard ===


@router.get("/", response_class=HTMLResponse)
async def dashboard(
    request: Request,
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Dashboard page."""
    entity_svc = EntityService(db)
    content_types = field_service.get_all_content_types()

    # Get stats for each content type
    stats = {}
    for ct_name in content_types.keys():
        stats[ct_name] = await entity_svc.count(ct_name)

    # Get recent posts with channel info
    from ..services.relation import RelationService

    relation_svc = RelationService(db)
    recent_posts = []
    posts = await entity_svc.find("post", limit=5, order_by="-created_at")
    for post in posts:
        data = entity_svc.serialize(post)
        # Get channel for this post
        related_channels = await relation_svc.get_related(post.id, "post_channel")
        if related_channels:
            channel_data = entity_svc.serialize(related_channels[0])
            data["channel_slug"] = channel_data.get("slug")
            data["channel_title"] = channel_data.get("title")
        recent_posts.append(data)

    # Check for updates
    from ..services.update import update_service

    update_info = await update_service.check_for_updates()

    context = await get_context(request, db, current_user, "dashboard")
    context.update(
        {
            "stats": stats,
            "recent_posts": recent_posts,
            "update_info": {
                "current_version": update_info.current_version,
                "latest_version": update_info.latest_version,
                "has_update": update_info.has_update,
                "release_url": update_info.release_url,
            },
        }
    )

    return templates.TemplateResponse("admin/dashboard.html", context)


# === Media ===


@router.get("/media", response_class=HTMLResponse)
async def media_list(
    request: Request,
    page: int = 1,
    q: str = "",
    type: str = "",
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Media library page."""
    require_feature("media")
    from ..services.media import MediaService

    media_svc = MediaService(db)

    per_page = get_settings().admin.per_page
    offset = (page - 1) * per_page

    # Prepare filters
    mime_type = f"{type}/" if type else None
    search = q if q else None

    items = await media_svc.find(limit=per_page, offset=offset, mime_type=mime_type, search=search)
    total = await media_svc.count(mime_type=mime_type, search=search)
    total_pages = (total + per_page - 1) // per_page if total > 0 else 0

    context = await get_context(request, db, current_user, "media")
    context.update(
        {
            "items": [media_svc.serialize(m) for m in items],
            "total": total,
            "page": page,
            "total_pages": total_pages,
            "message": request.query_params.get("message"),
            "search_query": q,
            "type_filter": type,
        }
    )

    return templates.TemplateResponse("admin/media.html", context)



"""Mail service for sending emails."""

import smtplib
from dataclasses import dataclass
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from ..config import settings


@dataclass
class EmailMessage:
    """Email message data."""

    to: str
    subject: str
    body: str
    html: str | None = None
    reply_to: str | None = None


class MailService:
    """
    Mail sending service.

    Supports SMTP.
    Configure via environment variables:
    - MAIL_HOST
    - MAIL_PORT
    - MAIL_USERNAME
    - MAIL_PASSWORD
    - MAIL_FROM
    - MAIL_FROM_NAME
    """

    def __init__(self):
        self.host = getattr(settings, "mail_host", None)
        self.port = getattr(settings, "mail_port", 587)
        self.username = getattr(settings, "mail_username", None)
        self.password = getattr(settings, "mail_password", None)
        self.from_email = getattr(settings, "mail_from", None)
        self.from_name = getattr(settings, "mail_from_name", "Focomy")
        self.use_tls = getattr(settings, "mail_use_tls", True)

    def is_configured(self) -> bool:
        """Check if mail is configured."""
        return bool(self.host and self.from_email)

    def send(self, message: EmailMessage) -> bool:
        """Send an email."""
        if not self.is_configured():
            print("Mail not configured, skipping send")
            return False

        try:
            msg = MIMEMultipart("alternative")
            msg["Subject"] = message.subject
            msg["From"] = f"{self.from_name} <{self.from_email}>"
            msg["To"] = message.to

            if message.reply_to:
                msg["Reply-To"] = message.reply_to

            # Add body
            msg.attach(MIMEText(message.body, "plain", "utf-8"))

            if message.html:
                msg.attach(MIMEText(message.html, "html", "utf-8"))

            # Send
            with smtplib.SMTP(self.host, self.port) as server:
                if self.use_tls:
                    server.starttls()
                if self.username and self.password:
                    server.login(self.username, self.password)
                server.sendmail(self.from_email, message.to, msg.as_string())

            return True

        except Exception as e:
            print(f"Failed to send email: {e}")
            return False

    def send_form_notification(
        self,
        to: str,
        form_title: str,
        submission_data: dict,
        reply_to: str | None = None,
    ) -> bool:
        """Send form submission notification."""
        # Build plain text body
        lines = [f"フォーム「{form_title}」から新しい送信がありました。", "", "--- 送信内容 ---"]
        for key, value in submission_data.items():
            lines.append(f"{key}: {value}")

        body = "\n".join(lines)

        # Build HTML body
        html_lines = [
            f"<h2>フォーム「{form_title}」から新しい送信がありました</h2>",
            "<table style='border-collapse: collapse; width: 100%;'>",
        ]
        for key, value in submission_data.items():
            html_lines.append(
                f"<tr><td style='border: 1px solid #ddd; padding: 8px; background: #f9f9f9;'><strong>{key}</strong></td>"
                f"<td style='border: 1px solid #ddd; padding: 8px;'>{value}</td></tr>"
            )
        html_lines.append("</table>")

        html = "\n".join(html_lines)

        return self.send(
            EmailMessage(
                to=to,
                subject=f"[Focomy] {form_title} - 新しい送信",
                body=body,
                html=html,
                reply_to=reply_to,
            )
        )


# Singleton
mail_service = MailService()



"""Migration Helpers - Safe database migration utilities.

Provides utilities for schema changes that need to be done carefully.
"""

from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession


class MigrationHelpers:
    """
    Utilities for safe database migrations.

    Usage:
        helpers = MigrationHelpers(db)

        # Add FK constraint safely
        await helpers.add_user_fk_constraints()

        # Verify integrity before constraint
        issues = await helpers.check_orphan_references()
    """

    def __init__(self, db: AsyncSession):
        self.db = db

    async def add_user_fk_constraints(self) -> dict:
        """
        Add FK constraints for created_by and updated_by columns.

        This adds self-referential FK constraints to the entities table.

        Returns:
            Migration result with status and any issues
        """
        result = {
            "success": False,
            "orphan_created_by": 0,
            "orphan_updated_by": 0,
            "nullified_count": 0,
            "constraints_added": [],
        }

        try:
            # First, check for orphan references
            orphans = await self.check_orphan_references()
            result["orphan_created_by"] = orphans["created_by_orphans"]
            result["orphan_updated_by"] = orphans["updated_by_orphans"]

            # Nullify orphan references (safe approach)
            if orphans["created_by_orphans"] > 0:
                await self.db.execute(
                    text(
                        """
                    UPDATE entities
                    SET created_by = NULL
                    WHERE created_by IS NOT NULL
                    AND created_by NOT IN (
                        SELECT id FROM entities WHERE type = 'user' AND deleted_at IS NULL
                    )
                """
                    )
                )
                result["nullified_count"] += orphans["created_by_orphans"]

            if orphans["updated_by_orphans"] > 0:
                await self.db.execute(
                    text(
                        """
                    UPDATE entities
                    SET updated_by = NULL
                    WHERE updated_by IS NOT NULL
                    AND updated_by NOT IN (
                        SELECT id FROM entities WHERE type = 'user' AND deleted_at IS NULL
                    )
                """
                    )
                )
                result["nullified_count"] += orphans["updated_by_orphans"]

            # Add FK constraints (PostgreSQL only)
            # Add created_by FK
            try:
                await self.db.execute(
                    text(
                        """
                    ALTER TABLE entities
                    ADD CONSTRAINT fk_entities_created_by
                    FOREIGN KEY (created_by) REFERENCES entities(id)
                    ON DELETE SET NULL
                """
                    )
                )
                result["constraints_added"].append("fk_entities_created_by")
            except Exception:
                pass  # Constraint might already exist

            # Add updated_by FK
            try:
                await self.db.execute(
                    text(
                        """
                    ALTER TABLE entities
                    ADD CONSTRAINT fk_entities_updated_by
                    FOREIGN KEY (updated_by) REFERENCES entities(id)
                    ON DELETE SET NULL
                """
                    )
                )
                result["constraints_added"].append("fk_entities_updated_by")
            except Exception:
                pass  # Constraint might already exist

            await self.db.commit()
            result["success"] = True

        except Exception as e:
            await self.db.rollback()
            result["error"] = str(e)

        return result

    async def check_orphan_references(self) -> dict:
        """
        Check for orphan created_by and updated_by references.

        Returns:
            Dict with counts of orphan references
        """
        result = {
            "created_by_orphans": 0,
            "updated_by_orphans": 0,
            "orphan_details": [],
        }

        # Check created_by orphans
        created_by_query = await self.db.execute(
            text(
                """
            SELECT COUNT(*) FROM entities e
            WHERE e.created_by IS NOT NULL
            AND NOT EXISTS (
                SELECT 1 FROM entities u
                WHERE u.id = e.created_by
                AND u.type = 'user'
            )
        """
            )
        )
        result["created_by_orphans"] = created_by_query.scalar() or 0

        # Check updated_by orphans
        updated_by_query = await self.db.execute(
            text(
                """
            SELECT COUNT(*) FROM entities e
            WHERE e.updated_by IS NOT NULL
            AND NOT EXISTS (
                SELECT 1 FROM entities u
                WHERE u.id = e.updated_by
                AND u.type = 'user'
            )
        """
            )
        )
        result["updated_by_orphans"] = updated_by_query.scalar() or 0

        return result

    async def remove_user_fk_constraints(self) -> dict:
        """Remove FK constraints (for rollback)."""
        result = {
            "success": False,
            "constraints_removed": [],
        }

        try:
            try:
                await self.db.execute(
                    text(
                        """
                    ALTER TABLE entities
                    DROP CONSTRAINT IF EXISTS fk_entities_created_by
                """
                    )
                )
                result["constraints_removed"].append("fk_entities_created_by")
            except Exception:
                pass

            try:
                await self.db.execute(
                    text(
                        """
                    ALTER TABLE entities
                    DROP CONSTRAINT IF EXISTS fk_entities_updated_by
                """
                    )
                )
                result["constraints_removed"].append("fk_entities_updated_by")
            except Exception:
                pass

            await self.db.commit()
            result["success"] = True

        except Exception as e:
            await self.db.rollback()
            result["error"] = str(e)

        return result

    async def add_indexed_fields_indexes(self, content_types: dict) -> dict:
        """
        Add database indexes for fields marked as indexed: true in YAML.

        Args:
            content_types: Dict of content type definitions

        Returns:
            Migration result
        """
        result = {
            "success": False,
            "indexes_created": [],
            "errors": [],
        }

        try:
            for type_name, type_def in content_types.items():
                fields = type_def.get("fields", [])

                for field in fields:
                    if field.get("indexed"):
                        field_name = field["name"]
                        index_name = f"idx_values_{type_name}_{field_name}"

                        try:
                            # Create partial index for this field
                            await self.db.execute(
                                text(
                                    f"""
                                CREATE INDEX IF NOT EXISTS {index_name}
                                ON entity_values (entity_id, value_text)
                                WHERE field_name = :field_name
                            """
                                ),
                                {"field_name": field_name},
                            )
                            result["indexes_created"].append(index_name)
                        except Exception as e:
                            result["errors"].append(f"{index_name}: {str(e)}")

            await self.db.commit()
            result["success"] = len(result["errors"]) == 0

        except Exception as e:
            await self.db.rollback()
            result["error"] = str(e)

        return result

    async def verify_schema_integrity(self) -> dict:
        """
        Verify database schema integrity.

        Checks:
        - Required tables exist
        - Required columns exist
        - Indexes exist
        """
        result = {
            "valid": True,
            "issues": [],
        }

        required_tables = ["entities", "entity_values", "relations"]

        for table in required_tables:
            try:
                check = await self.db.execute(
                    text(
                        """
                    SELECT EXISTS (
                        SELECT FROM information_schema.tables
                        WHERE table_name = :table
                    )
                """
                    ),
                    {"table": table},
                )

                exists = check.scalar()
                if not exists:
                    result["valid"] = False
                    result["issues"].append(f"Missing table: {table}")

            except Exception as e:
                result["issues"].append(f"Error checking table {table}: {str(e)}")

        return result


def get_migration_helpers(db: AsyncSession) -> MigrationHelpers:
    return MigrationHelpers(db)



"""Diff detection and differential import operations."""

from __future__ import annotations

import logging

from ...models import ImportJobPhase, ImportJobStatus
from ...utils import utcnow
from .constants import WP_STATUS_MAP
from .wxr_parser import WXRData

logger = logging.getLogger(__name__)


class DiffMixin:
    """Diff detection and differential import methods.

    Mixed into WordPressImportService.
    Requires: self.db, self.entity_svc, self.sanitizer,
              self.update_job(), self.get_job(), self._fetch_source_data(),
              self._find_by_wp_id()
    """

    async def detect_diff(self, job_id: str) -> dict | None:
        """
        Detect differences between WordPress data and database.

        Returns a summary of new/updated/unchanged/deleted items.
        """
        job = await self.get_job(job_id)
        if not job:
            return None

        try:
            await self.update_job(
                job_id,
                status=ImportJobStatus.ANALYZING,
                phase=ImportJobPhase.ANALYZE,
                progress_message="Detecting differences...",
            )

            # Get WXR data from source
            wxr_data = await self._fetch_source_data(job)
            if not wxr_data:
                raise ValueError("Failed to fetch source data")

            # Run diff detection
            from .diff_detector import DiffDetector

            detector = DiffDetector(self.db)
            diff_result = await detector.detect(wxr_data)

            # Store diff result in job config
            config = job.config or {}
            config["diff_result"] = diff_result.to_dict()
            await self.update_job(
                job_id,
                status=ImportJobStatus.PENDING,
                phase=ImportJobPhase.INIT,
                config=config,
                progress_message="Diff detection complete",
            )

            return {
                "success": True,
                **diff_result.to_dict(),
            }

        except Exception as e:
            logger.exception(f"Diff detection failed for job {job_id}")
            return {
                "success": False,
                "error": str(e),
            }

    async def import_diff(
        self,
        job_id: str,
        import_new: bool = True,
        import_updated: bool = True,
        delete_removed: bool = False,
    ) -> dict | None:
        """
        Import only the differences (new and updated items).

        Args:
            job_id: Import job ID
            import_new: Import new items
            import_updated: Import updated items (will update existing)
            delete_removed: Delete items that are no longer in WordPress
        """
        job = await self.get_job(job_id)
        if not job:
            return None

        config = job.config or {}
        diff_data = config.get("diff_result")

        if not diff_data:
            return {
                "success": False,
                "error": "No diff result found. Run detect_diff first.",
            }

        try:
            await self.update_job(
                job_id,
                status=ImportJobStatus.IMPORTING,
                phase=ImportJobPhase.POSTS,
                started_at=utcnow(),
                progress_message="Starting diff import...",
            )

            # Get WXR data again
            wxr_data = await self._fetch_source_data(job)
            if not wxr_data:
                raise ValueError("Failed to fetch source data")

            results = {
                "new_imported": 0,
                "updated": 0,
                "deleted": 0,
                "errors": [],
            }

            # Import new items
            if import_new:
                await self._import_new_from_diff(job_id, wxr_data, diff_data, results)

            # Update existing items
            if import_updated:
                await self._update_from_diff(job_id, wxr_data, diff_data, results)

            # Delete removed items
            if delete_removed:
                await self._delete_from_diff(job_id, diff_data, results)

            await self.update_job(
                job_id,
                status=ImportJobStatus.COMPLETED,
                phase=ImportJobPhase.COMPLETE,
                completed_at=utcnow(),
                progress_message="Diff import complete",
            )

            return {
                "success": True,
                **results,
            }

        except Exception as e:
            logger.exception(f"Diff import failed for job {job_id}")
            await self.update_job(
                job_id,
                status=ImportJobStatus.FAILED,
                completed_at=utcnow(),
                errors=[str(e)],
                progress_message=f"Diff import failed: {str(e)}",
            )
            return {
                "success": False,
                "error": str(e),
            }

    async def _import_new_from_diff(
        self,
        job_id: str,
        wxr_data: WXRData,
        diff_data: dict,
        results: dict,
    ) -> None:
        """Import new items from diff."""
        # Build lookup maps for WXR data
        posts_map = {p.id: p for p in wxr_data.posts if p.post_type == "post"}
        pages_map = {p.id: p for p in wxr_data.posts if p.post_type == "page"}
        cats_map = {c.id: c for c in wxr_data.categories}
        tags_map = {t.id: t for t in wxr_data.tags}

        # Import new posts
        for item in diff_data.get("new", {}).get("posts", []):
            wp_id = item["wp_id"]
            post = posts_map.get(wp_id)
            if post:
                try:
                    # Sanitize content
                    content_result = self.sanitizer.sanitize(post.content or "")
                    excerpt_result = self.sanitizer.sanitize(post.excerpt or "")
                    if content_result.had_issues or excerpt_result.had_issues:
                        logger.warning(f"Sanitized content in new post {wp_id}")

                    await self.entity_svc.create("post", {
                        "title": post.title,
                        "slug": post.slug,
                        "content": content_result.content,
                        "excerpt": excerpt_result.content,
                        "status": WP_STATUS_MAP.get(post.status, "draft"),
                        "wp_id": post.id,
                        "wp_modified": post.modified_at.isoformat() if post.modified_at else None,
                    })
                    results["new_imported"] += 1
                except Exception as e:
                    results["errors"].append(f"Post {wp_id}: {e}")

        # Import new pages
        for item in diff_data.get("new", {}).get("pages", []):
            wp_id = item["wp_id"]
            page = pages_map.get(wp_id)
            if page:
                try:
                    # Sanitize content
                    content_result = self.sanitizer.sanitize(page.content or "")
                    if content_result.had_issues:
                        logger.warning(f"Sanitized content in new page {wp_id}")

                    await self.entity_svc.create("page", {
                        "title": page.title,
                        "slug": page.slug,
                        "content": content_result.content,
                        "status": WP_STATUS_MAP.get(page.status, "draft"),
                        "wp_id": page.id,
                        "wp_modified": page.modified_at.isoformat() if page.modified_at else None,
                    })
                    results["new_imported"] += 1
                except Exception as e:
                    results["errors"].append(f"Page {wp_id}: {e}")

        # Import new categories
        for item in diff_data.get("new", {}).get("categories", []):
            wp_id = item["wp_id"]
            cat = cats_map.get(wp_id)
            if cat:
                try:
                    await self.entity_svc.create("category", {
                        "name": cat.name,
                        "slug": cat.slug,
                        "wp_id": cat.id,
                    })
                    results["new_imported"] += 1
                except Exception as e:
                    results["errors"].append(f"Category {wp_id}: {e}")

        # Import new tags
        for item in diff_data.get("new", {}).get("tags", []):
            wp_id = item["wp_id"]
            tag = tags_map.get(wp_id)
            if tag:
                try:
                    await self.entity_svc.create("tag", {
                        "name": tag.name,
                        "slug": tag.slug,
                        "wp_id": tag.id,
                    })
                    results["new_imported"] += 1
                except Exception as e:
                    results["errors"].append(f"Tag {wp_id}: {e}")

        await self.update_job(
            job_id,
            progress_message=f"Imported {results['new_imported']} new items",
        )

    async def _update_from_diff(
        self,
        job_id: str,
        wxr_data: WXRData,
        diff_data: dict,
        results: dict,
    ) -> None:
        """Update existing items from diff."""
        posts_map = {p.id: p for p in wxr_data.posts if p.post_type == "post"}
        pages_map = {p.id: p for p in wxr_data.posts if p.post_type == "page"}
        cats_map = {c.id: c for c in wxr_data.categories}
        tags_map = {t.id: t for t in wxr_data.tags}

        # Update posts
        for item in diff_data.get("updated", {}).get("posts", []):
            wp_id = item["wp_id"]
            post = posts_map.get(wp_id)
            if post:
                try:
                    existing = await self._find_by_wp_id("post", wp_id)
                    if existing:
                        # Sanitize content
                        content_result = self.sanitizer.sanitize(post.content or "")
                        excerpt_result = self.sanitizer.sanitize(post.excerpt or "")
                        if content_result.had_issues or excerpt_result.had_issues:
                            logger.warning(f"Sanitized content in updated post {wp_id}")

                        await self.entity_svc.update("post", existing.id, {
                            "title": post.title,
                            "slug": post.slug,
                            "content": content_result.content,
                            "excerpt": excerpt_result.content,
                            "status": WP_STATUS_MAP.get(post.status, "draft"),
                            "wp_modified": post.modified_at.isoformat() if post.modified_at else None,
                        })
                        results["updated"] += 1
                except Exception as e:
                    results["errors"].append(f"Update post {wp_id}: {e}")

        # Update pages
        for item in diff_data.get("updated", {}).get("pages", []):
            wp_id = item["wp_id"]
            page = pages_map.get(wp_id)
            if page:
                try:
                    existing = await self._find_by_wp_id("page", wp_id)
                    if existing:
                        # Sanitize content
                        content_result = self.sanitizer.sanitize(page.content or "")
                        if content_result.had_issues:
                            logger.warning(f"Sanitized content in updated page {wp_id}")

                        await self.entity_svc.update("page", existing.id, {
                            "title": page.title,
                            "slug": page.slug,
                            "content": content_result.content,
                            "status": WP_STATUS_MAP.get(page.status, "draft"),
                            "wp_modified": page.modified_at.isoformat() if page.modified_at else None,
                        })
                        results["updated"] += 1
                except Exception as e:
                    results["errors"].append(f"Update page {wp_id}: {e}")

        # Update categories
        for item in diff_data.get("updated", {}).get("categories", []):
            wp_id = item["wp_id"]
            cat = cats_map.get(wp_id)
            if cat:
                try:
                    existing = await self._find_by_wp_id("category", wp_id)
                    if existing:
                        await self.entity_svc.update("category", existing.id, {
                            "name": cat.name,
                            "slug": cat.slug,
                        })
                        results["updated"] += 1
                except Exception as e:
                    results["errors"].append(f"Update category {wp_id}: {e}")

        # Update tags
        for item in diff_data.get("updated", {}).get("tags", []):
            wp_id = item["wp_id"]
            tag = tags_map.get(wp_id)
            if tag:
                try:
                    existing = await self._find_by_wp_id("tag", wp_id)
                    if existing:
                        await self.entity_svc.update("tag", existing.id, {
                            "name": tag.name,
                            "slug": tag.slug,
                        })
                        results["updated"] += 1
                except Exception as e:
                    results["errors"].append(f"Update tag {wp_id}: {e}")

        await self.update_job(
            job_id,
            progress_message=f"Updated {results['updated']} items",
        )

    async def _delete_from_diff(
        self,
        job_id: str,
        diff_data: dict,
        results: dict,
    ) -> None:
        """Delete items that were removed from WordPress."""
        entity_type_map = {
            "posts": "post",
            "pages": "page",
            "media": "media",
            "categories": "category",
            "tags": "tag",
            "authors": "user",
        }

        for diff_type, entity_type in entity_type_map.items():
            for item in diff_data.get("deleted", {}).get(diff_type, []):
                wp_id = item["wp_id"]
                try:
                    existing = await self._find_by_wp_id(entity_type, wp_id)
                    if existing:
                        await self.entity_svc.delete(entity_type, existing.id)
                        results["deleted"] += 1
                except Exception as e:
                    results["errors"].append(f"Delete {diff_type} {wp_id}: {e}")

        await self.update_job(
            job_id,
            progress_message=f"Deleted {results['deleted']} items",
        )



"""Cache Service - Supports Redis and in-memory caching.

Provides a unified caching interface that:
- Uses Redis when configured
- Falls back to in-memory cache otherwise
- Supports TTL (time-to-live)
- Supports cache invalidation patterns
"""

import asyncio
import hashlib
import json
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import datetime, timedelta, timezone


def utcnow() -> datetime:
    """Return current UTC time as timezone-aware datetime."""
    return datetime.now(timezone.utc)
from functools import wraps
from typing import Any, Optional, TypeVar

try:
    import redis.asyncio as aioredis

    REDIS_AVAILABLE = True
except ImportError:
    REDIS_AVAILABLE = False


T = TypeVar("T")

logger = logging.getLogger(__name__)


class CacheBackend(ABC):
    """Abstract base class for cache backends."""

    @abstractmethod
    async def get(self, key: str) -> Any | None:
        pass

    @abstractmethod
    async def set(self, key: str, value: Any, ttl: int | None = None) -> None:
        pass

    @abstractmethod
    async def delete(self, key: str) -> bool:
        pass

    @abstractmethod
    async def delete_pattern(self, pattern: str) -> int:
        pass

    @abstractmethod
    async def clear(self) -> int:
        pass

    @abstractmethod
    async def exists(self, key: str) -> bool:
        pass


class CacheEntry:
    """A single cache entry with expiration."""

    def __init__(self, value: Any, ttl_seconds: int):
        self.value = value
        self.expires_at = utcnow() + timedelta(seconds=ttl_seconds)

    def is_expired(self) -> bool:
        return utcnow() > self.expires_at


class InMemoryBackend(CacheBackend):
    """In-memory cache backend with TTL support."""

    def __init__(self):
        self._cache: dict[str, CacheEntry] = {}
        self._lock = asyncio.Lock()

    async def get(self, key: str) -> Any | None:
        async with self._lock:
            entry = self._cache.get(key)
            if entry is None:
                return None
            if entry.is_expired():
                del self._cache[key]
                return None
            return entry.value

    async def set(self, key: str, value: Any, ttl: int | None = None) -> None:
        ttl = ttl or 300  # Default 5 minutes
        async with self._lock:
            self._cache[key] = CacheEntry(value, ttl)

    async def delete(self, key: str) -> bool:
        async with self._lock:
            if key in self._cache:
                del self._cache[key]
                return True
            return False

    async def delete_pattern(self, pattern: str) -> int:
        """Delete keys matching glob pattern."""
        import fnmatch

        async with self._lock:
            to_delete = [k for k in self._cache.keys() if fnmatch.fnmatch(k, pattern)]
            for key in to_delete:
                del self._cache[key]
            return len(to_delete)

    async def clear(self) -> int:
        async with self._lock:
            count = len(self._cache)
            self._cache.clear()
            return count

    async def exists(self, key: str) -> bool:
        async with self._lock:
            entry = self._cache.get(key)
            if entry is None:
                return False
            if entry.is_expired():
                del self._cache[key]
                return False
            return True

    def cleanup_expired(self) -> int:
        """Remove expired entries (sync for periodic cleanup)."""
        expired = [k for k, v in self._cache.items() if v.is_expired()]
        for key in expired:
            del self._cache[key]
        return len(expired)

    def stats(self) -> dict:
        """Get cache statistics."""
        valid = sum(1 for v in self._cache.values() if not v.is_expired())
        return {
            "total_entries": len(self._cache),
            "valid_entries": valid,
            "expired_entries": len(self._cache) - valid,
            "backend": "memory",
        }


class RedisBackend(CacheBackend):
    """Redis cache backend."""

    def __init__(self, client, prefix: str = "focomy:"):
        self._redis = client
        self._prefix = prefix

    def _key(self, key: str) -> str:
        return f"{self._prefix}{key}"

    async def get(self, key: str) -> Any | None:
        try:
            data = await self._redis.get(self._key(key))
            if data is None:
                return None
            return json.loads(data)
        except Exception:
            return None

    async def set(self, key: str, value: Any, ttl: int | None = None) -> None:
        ttl = ttl or 300
        try:
            serialized = json.dumps(value, default=str)
            await self._redis.setex(self._key(key), ttl, serialized)
        except Exception as e:
            logger.debug(f"Redis cache set failed for key {key}: {e}")

    async def delete(self, key: str) -> bool:
        try:
            result = await self._redis.delete(self._key(key))
            return result > 0
        except Exception:
            return False

    async def delete_pattern(self, pattern: str) -> int:
        """Delete keys matching pattern using SCAN."""
        try:
            count = 0
            cursor = 0
            match_pattern = self._key(pattern)

            while True:
                cursor, keys = await self._redis.scan(cursor, match=match_pattern, count=100)
                if keys:
                    await self._redis.delete(*keys)
                    count += len(keys)
                if cursor == 0:
                    break
            return count
        except Exception:
            return 0

    async def clear(self) -> int:
        """Clear all focomy cache entries."""
        return await self.delete_pattern("*")

    async def exists(self, key: str) -> bool:
        try:
            return bool(await self._redis.exists(self._key(key)))
        except Exception:
            return False

    def stats(self) -> dict:
        """Get cache statistics."""
        return {"backend": "redis"}


class CacheService:
    """
    Unified cache service with Redis support.

    Usage:
        cache = CacheService()
        await cache.connect()  # Optional: auto-connects on first use

        # Basic operations
        await cache.set("key", {"data": "value"}, ttl=300)
        value = await cache.get("key")
        await cache.delete("key")

        # Pattern invalidation
        await cache.invalidate_pattern("entity:*")
    """

    _instance: Optional["CacheService"] = None

    def __new__(cls, redis_url: str | None = None):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._redis_url = redis_url
            cls._instance._backend = None
            cls._instance._connected = False
        return cls._instance

    async def connect(self, redis_url: str | None = None) -> None:
        """Initialize cache backend."""
        if self._connected:
            return

        url = redis_url or self._redis_url

        # Try Redis if available and configured
        if REDIS_AVAILABLE and url:
            try:
                client = aioredis.from_url(url, decode_responses=True)
                await client.ping()
                self._backend = RedisBackend(client)
                self._connected = True
                return
            except Exception as e:
                logger.debug(f"Redis connection failed: {e}")

        # Fall back to in-memory
        self._backend = InMemoryBackend()
        self._connected = True

    async def disconnect(self) -> None:
        """Close connections."""
        if isinstance(self._backend, RedisBackend):
            await self._backend._redis.close()
        self._backend = None
        self._connected = False

    @property
    def backend(self) -> CacheBackend:
        if self._backend is None:
            self._backend = InMemoryBackend()
            self._connected = True
        return self._backend

    # Core operations (async)
    async def get(self, key: str) -> Any | None:
        return await self.backend.get(key)

    async def set(self, key: str, value: Any, ttl: int | None = None) -> None:
        await self.backend.set(key, value, ttl)

    async def delete(self, key: str) -> bool:
        return await self.backend.delete(key)

    async def invalidate_pattern(self, pattern: str) -> int:
        return await self.backend.delete_pattern(pattern)

    async def invalidate_all(self) -> int:
        return await self.backend.clear()

    async def exists(self, key: str) -> bool:
        return await self.backend.exists(key)

    # Sync operations (for backward compatibility)
    def get_sync(self, key: str) -> Any | None:
        """Synchronous get (only works with in-memory backend)."""
        if isinstance(self.backend, InMemoryBackend):
            entry = self.backend._cache.get(key)
            if entry and not entry.is_expired():
                return entry.value
        return None

    def set_sync(self, key: str, value: Any, ttl: int = 300) -> None:
        """Synchronous set (only works with in-memory backend)."""
        if isinstance(self.backend, InMemoryBackend):
            self.backend._cache[key] = CacheEntry(value, ttl)

    def delete_sync(self, key: str) -> bool:
        """Synchronous delete (only works with in-memory backend)."""
        if isinstance(self.backend, InMemoryBackend):
            if key in self.backend._cache:
                del self.backend._cache[key]
                return True
        return False

    # Utility methods
    async def get_or_set(
        self,
        key: str,
        factory: Callable[[], Any],
        ttl: int | None = None,
    ) -> Any:
        """Get from cache or compute and store."""
        value = await self.get(key)
        if value is not None:
            return value

        if asyncio.iscoroutinefunction(factory):
            value = await factory()
        else:
            value = factory()

        if value is not None:
            await self.set(key, value, ttl)
        return value

    def stats(self) -> dict:
        """Get cache statistics."""
        if isinstance(self.backend, InMemoryBackend):
            return self.backend.stats()
        elif isinstance(self.backend, RedisBackend):
            return self.backend.stats()
        return {"backend": "none"}


# Singleton instance
cache_service = CacheService()


def make_cache_key(*args, **kwargs) -> str:
    """Generate a cache key from arguments."""
    key_parts = [str(a) for a in args]
    key_parts.extend(f"{k}={v}" for k, v in sorted(kwargs.items()))
    key_str = ":".join(key_parts)
    return hashlib.md5(key_str.encode()).hexdigest()


def cached(prefix: str, ttl: int = 300):
    """
    Decorator for caching async function results.

    Usage:
        @cached("page", ttl=300)
        async def get_page(slug):
            ...
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            key = f"{prefix}:{make_cache_key(*args, **kwargs)}"

            cached_value = await cache_service.get(key)
            if cached_value is not None:
                return cached_value

            result = await func(*args, **kwargs)

            if result is not None:
                await cache_service.set(key, result, ttl)

            return result

        # Add invalidation helpers
        wrapper.cache_key = lambda *a, **kw: f"{prefix}:{make_cache_key(*a, **kw)}"
        wrapper.invalidate = lambda *a, **kw: cache_service.delete(
            f"{prefix}:{make_cache_key(*a, **kw)}"
        )
        wrapper.invalidate_all = lambda: cache_service.invalidate_pattern(f"{prefix}:*")

        return wrapper

    return decorator


def invalidate_cache(prefix: str) -> int:
    """Invalidate all cache entries with a given prefix (sync, for backward compat)."""
    if isinstance(cache_service.backend, InMemoryBackend):
        to_delete = [k for k in cache_service.backend._cache.keys() if k.startswith(prefix)]
        for key in to_delete:
            del cache_service.backend._cache[key]
        return len(to_delete)
    return 0


# Cache key helpers
class CacheKeys:
    """Common cache key patterns."""

    @staticmethod
    def entity(entity_id: str) -> str:
        return f"entity:{entity_id}"

    @staticmethod
    def entity_list(entity_type: str, page: int = 1) -> str:
        return f"list:{entity_type}:{page}"

    @staticmethod
    def settings(key: str = "all") -> str:
        return f"settings:{key}"

    @staticmethod
    def theme(theme_name: str) -> str:
        return f"theme:{theme_name}"



"""Pagination Service - Standardized pagination helpers.

Provides both offset-based and cursor-based pagination.
"""

import base64
import json
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from pydantic import BaseModel

T = TypeVar("T")


@dataclass
class PageInfo:
    """Pagination metadata."""

    page: int
    per_page: int
    total: int
    total_pages: int
    has_next: bool
    has_prev: bool


class PaginatedResult(BaseModel, Generic[T]):
    """Standard paginated response."""

    items: list
    page: int
    per_page: int
    total: int
    total_pages: int
    has_next: bool
    has_prev: bool

    @classmethod
    def create(
        cls,
        items: list,
        page: int,
        per_page: int,
        total: int,
    ) -> "PaginatedResult":
        total_pages = (total + per_page - 1) // per_page if per_page > 0 else 0
        return cls(
            items=items,
            page=page,
            per_page=per_page,
            total=total,
            total_pages=total_pages,
            has_next=page < total_pages,
            has_prev=page > 1,
        )

    def to_dict(self) -> dict:
        return {
            "items": self.items,
            "pagination": {
                "page": self.page,
                "per_page": self.per_page,
                "total": self.total,
                "total_pages": self.total_pages,
                "has_next": self.has_next,
                "has_prev": self.has_prev,
            },
        }


@dataclass
class CursorInfo:
    """Cursor-based pagination metadata."""

    cursor: str | None
    limit: int
    has_more: bool
    next_cursor: str | None


class CursorPaginatedResult(BaseModel, Generic[T]):
    """Cursor-based paginated response."""

    items: list
    limit: int
    has_more: bool
    next_cursor: str | None

    @classmethod
    def create(
        cls,
        items: list,
        limit: int,
        cursor_field: str = "id",
    ) -> "CursorPaginatedResult":
        has_more = len(items) > limit
        if has_more:
            items = items[:limit]

        next_cursor = None
        if items and has_more:
            last_item = items[-1]
            if isinstance(last_item, dict):
                cursor_value = last_item.get(cursor_field)
            else:
                cursor_value = getattr(last_item, cursor_field, None)
            if cursor_value:
                next_cursor = encode_cursor(cursor_value)

        return cls(
            items=items,
            limit=limit,
            has_more=has_more,
            next_cursor=next_cursor,
        )

    def to_dict(self) -> dict:
        return {
            "items": self.items,
            "pagination": {
                "limit": self.limit,
                "has_more": self.has_more,
                "next_cursor": self.next_cursor,
            },
        }


def encode_cursor(value: Any) -> str:
    """Encode a cursor value."""
    data = json.dumps({"v": str(value)})
    return base64.urlsafe_b64encode(data.encode()).decode()


def decode_cursor(cursor: str) -> str | None:
    """Decode a cursor value."""
    try:
        data = base64.urlsafe_b64decode(cursor.encode())
        parsed = json.loads(data.decode())
        return parsed.get("v")
    except Exception:
        return None


class PaginationParams(BaseModel):
    """Standard pagination parameters."""

    page: int = 1
    per_page: int = 20
    cursor: str | None = None
    sort: str = "-created_at"

    @property
    def offset(self) -> int:
        return (self.page - 1) * self.per_page

    @property
    def decoded_cursor(self) -> str | None:
        if self.cursor:
            return decode_cursor(self.cursor)
        return None


def calculate_pages(total: int, per_page: int) -> int:
    """Calculate total pages."""
    if per_page <= 0:
        return 0
    return (total + per_page - 1) // per_page


def get_page_range(current: int, total_pages: int, window: int = 2) -> list[int]:
    """Get page numbers to display in pagination UI."""
    start = max(1, current - window)
    end = min(total_pages, current + window)

    pages = list(range(start, end + 1))

    # Add first page if not included
    if start > 1:
        pages = [1, ...] + pages if start > 2 else [1] + pages

    # Add last page if not included
    if end < total_pages:
        pages = pages + [..., total_pages] if end < total_pages - 1 else pages + [total_pages]

    return pages



"""WorkflowService - approval workflow management."""

from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path

import yaml
from sqlalchemy.ext.asyncio import AsyncSession

from ..config import settings
from ..utils import utcnow
from .entity import EntityService


@dataclass
class WorkflowState:
    """Workflow state definition."""

    name: str
    label: str
    color: str = "gray"
    initial: bool = False
    final: bool = False


@dataclass
class WorkflowTransition:
    """Workflow transition definition."""

    from_state: str
    to_state: str
    label: str
    permissions: list[str] = field(default_factory=list)
    require_comment: bool = False
    actions: list[dict] = field(default_factory=list)


@dataclass
class WorkflowDefinition:
    """Complete workflow definition."""

    name: str
    label: str
    description: str = ""
    states: list[WorkflowState] = field(default_factory=list)
    transitions: list[WorkflowTransition] = field(default_factory=list)

    @property
    def initial_state(self) -> str | None:
        """Get the initial state name."""
        for state in self.states:
            if state.initial:
                return state.name
        return self.states[0].name if self.states else None

    def get_state(self, name: str) -> WorkflowState | None:
        """Get state by name."""
        for state in self.states:
            if state.name == name:
                return state
        return None

    def get_transition(self, from_state: str, to_state: str) -> WorkflowTransition | None:
        """Get transition between states."""
        for t in self.transitions:
            if t.from_state == from_state and t.to_state == to_state:
                return t
        return None

    def get_available_transitions(self, current_state: str) -> list[WorkflowTransition]:
        """Get all transitions from current state."""
        return [t for t in self.transitions if t.from_state == current_state]


@dataclass
class WorkflowHistoryEntry:
    """Workflow history entry."""

    entity_id: str
    from_state: str
    to_state: str
    user_id: str
    comment: str | None
    created_at: datetime


class WorkflowService:
    """
    Workflow management service.

    Handles state transitions, permission checks, and history tracking.
    """

    def __init__(self, db: AsyncSession):
        self.db = db
        self.entity_svc = EntityService(db)
        self._workflows: dict[str, WorkflowDefinition] = {}
        self._loaded = False

    async def _load_workflows(self) -> None:
        """Load workflow definitions from YAML files."""
        if self._loaded:
            return

        workflows_dir = settings.base_dir / "workflows"
        if workflows_dir.exists():
            for path in workflows_dir.glob("*.yaml"):
                wf = self._load_workflow(path)
                if wf:
                    self._workflows[wf.name] = wf

        self._loaded = True

    def _load_workflow(self, path: Path) -> WorkflowDefinition | None:
        """Load a single workflow from YAML."""
        try:
            with open(path, encoding="utf-8") as f:
                data = yaml.safe_load(f)

            if not data:
                return None

            states = []
            for s in data.get("states", []):
                states.append(
                    WorkflowState(
                        name=s.get("name", ""),
                        label=s.get("label", s.get("name", "")),
                        color=s.get("color", "gray"),
                        initial=s.get("initial", False),
                        final=s.get("final", False),
                    )
                )

            transitions = []
            for t in data.get("transitions", []):
                transitions.append(
                    WorkflowTransition(
                        from_state=t.get("from", ""),
                        to_state=t.get("to", ""),
                        label=t.get("label", ""),
                        permissions=t.get("permissions", []),
                        require_comment=t.get("require_comment", False),
                        actions=t.get("actions", []),
                    )
                )

            return WorkflowDefinition(
                name=data.get("name", path.stem),
                label=data.get("label", data.get("name", path.stem)),
                description=data.get("description", ""),
                states=states,
                transitions=transitions,
            )

        except Exception as e:
            print(f"Error loading workflow {path}: {e}")
            return None

    async def get_workflow(self, name: str) -> WorkflowDefinition | None:
        """Get workflow definition by name."""
        await self._load_workflows()
        return self._workflows.get(name)

    async def get_all_workflows(self) -> dict[str, WorkflowDefinition]:
        """Get all workflow definitions."""
        await self._load_workflows()
        return self._workflows.copy()

    async def get_entity_state(self, entity_id: str) -> str | None:
        """Get current workflow state of an entity."""
        entity = await self.entity_svc.get(entity_id)
        if not entity:
            return None

        data = self.entity_svc.serialize(entity)
        return data.get("workflow_state")

    async def transition(
        self,
        entity_id: str,
        to_state: str,
        user_id: str,
        comment: str = None,
        user_roles: list[str] = None,
    ) -> tuple[bool, str]:
        """
        Execute a workflow transition.

        Args:
            entity_id: Entity to transition
            to_state: Target state
            user_id: User performing the transition
            comment: Optional comment
            user_roles: User's roles for permission check

        Returns:
            Tuple of (success, error_message)
        """
        entity = await self.entity_svc.get(entity_id)
        if not entity:
            return False, "Entity not found"

        data = self.entity_svc.serialize(entity)
        content_type = entity.type

        # Get workflow for content type
        # (You'd typically store workflow name on the content type definition)
        workflow_name = data.get("_workflow")
        if not workflow_name:
            # Default workflow based on content type
            workflow_name = f"{content_type}_workflow"

        workflow = await self.get_workflow(workflow_name)
        if not workflow:
            # Try generic content review workflow
            workflow = await self.get_workflow("content_review")
            if not workflow:
                return False, "Workflow not found"

        # Get current state
        current_state = data.get("workflow_state")
        if not current_state:
            current_state = workflow.initial_state

        # Find transition
        transition = workflow.get_transition(current_state, to_state)
        if not transition:
            return False, f"Cannot transition from {current_state} to {to_state}"

        # Check permissions
        if transition.permissions and user_roles:
            has_permission = any(r in transition.permissions for r in user_roles)
            if not has_permission:
                return False, "Permission denied for this transition"

        # Check required comment
        if transition.require_comment and not comment:
            return False, "Comment is required for this transition"

        # Execute transition
        await self.entity_svc.update(
            entity_id,
            {"workflow_state": to_state},
            user_id=user_id,
        )

        # Record history
        await self._record_history(entity_id, current_state, to_state, user_id, comment)

        # Execute actions
        for action in transition.actions:
            await self._execute_action(action, entity_id, user_id, comment)

        return True, ""

    async def get_available_transitions(
        self,
        entity_id: str,
        user_roles: list[str] = None,
    ) -> list[WorkflowTransition]:
        """
        Get available transitions for an entity.

        Args:
            entity_id: Entity ID
            user_roles: User's roles for filtering

        Returns:
            List of available transitions
        """
        entity = await self.entity_svc.get(entity_id)
        if not entity:
            return []

        data = self.entity_svc.serialize(entity)
        content_type = entity.type

        # Get workflow
        workflow_name = data.get("_workflow", f"{content_type}_workflow")
        workflow = await self.get_workflow(workflow_name)
        if not workflow:
            workflow = await self.get_workflow("content_review")
            if not workflow:
                return []

        current_state = data.get("workflow_state", workflow.initial_state)
        all_transitions = workflow.get_available_transitions(current_state)

        # Filter by permissions
        if user_roles:
            return [
                t
                for t in all_transitions
                if not t.permissions or any(r in t.permissions for r in user_roles)
            ]

        return all_transitions

    async def get_history(
        self,
        entity_id: str,
        limit: int = 50,
    ) -> list[dict]:
        """Get workflow history for an entity."""
        # Query from workflow_history content type
        histories = await self.entity_svc.find(
            "workflow_history",
            limit=limit,
            order_by="-created_at",
            filters={"entity_id": entity_id},
        )

        return [self.entity_svc.serialize(h) for h in histories]

    async def _record_history(
        self,
        entity_id: str,
        from_state: str,
        to_state: str,
        user_id: str,
        comment: str = None,
    ) -> None:
        """Record a workflow transition in history."""
        # Create workflow_history entity
        await self.entity_svc.create(
            "workflow_history",
            {
                "entity_id": entity_id,
                "from_state": from_state,
                "to_state": to_state,
                "comment": comment or "",
            },
            user_id=user_id,
        )

    async def _execute_action(
        self,
        action: dict,
        entity_id: str,
        user_id: str,
        comment: str = None,
    ) -> None:
        """Execute a workflow action."""
        action_type = action.get("type")

        if action_type == "set_field":
            # Set a field value
            field_name = action.get("field")
            value = action.get("value")

            if value == "now":
                value = utcnow().isoformat()

            await self.entity_svc.update(
                entity_id,
                {field_name: value},
                user_id=user_id,
            )

        elif action_type == "notify":
            # Send notification (placeholder - integrate with notification service)
            action.get("to")
            message = action.get("message", "")
            if comment:
                message = message.replace("{comment}", comment)
            # TODO: Integrate with notification service
            pass

        elif action_type == "webhook":
            # Call webhook (placeholder)
            action.get("url")
            # TODO: Implement webhook call
            pass


# Default workflow for content
DEFAULT_CONTENT_WORKFLOW = """
name: content_review
label: コンテンツレビュー
description: 投稿の承認フロー

states:
  - name: draft
    label: 下書き
    initial: true
    color: gray

  - name: pending_review
    label: レビュー待ち
    color: yellow

  - name: in_review
    label: レビュー中
    color: blue

  - name: approved
    label: 承認済み
    color: green

  - name: rejected
    label: 却下
    color: red

  - name: published
    label: 公開中
    final: true
    color: green

transitions:
  - from: draft
    to: pending_review
    label: レビュー依頼
    permissions: [author, editor, admin]

  - from: pending_review
    to: in_review
    label: レビュー開始
    permissions: [editor, admin]

  - from: in_review
    to: approved
    label: 承認
    permissions: [editor, admin]
    actions:
      - type: notify
        to: author
        message: 投稿が承認されました

  - from: in_review
    to: rejected
    label: 却下
    permissions: [editor, admin]
    require_comment: true
    actions:
      - type: notify
        to: author
        message: "投稿が却下されました: {comment}"

  - from: rejected
    to: draft
    label: 修正
    permissions: [author, editor, admin]

  - from: approved
    to: published
    label: 公開
    permissions: [editor, admin]
    actions:
      - type: set_field
        field: published_at
        value: now

  - from: published
    to: draft
    label: 非公開に戻す
    permissions: [admin]
"""



"""
HTML Sanitizer Service for Editor.js blocks and user-generated content.

Prevents XSS attacks by sanitizing HTML output.
"""

import html
import re
from typing import Any

# Allowed HTML tags and their attributes
ALLOWED_TAGS = {
    # Headings
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    # Text formatting
    "p",
    "br",
    "hr",
    "b",
    "strong",
    "i",
    "em",
    "u",
    "s",
    "strike",
    "del",
    "sup",
    "sub",
    "mark",
    "small",
    # Lists
    "ul",
    "ol",
    "li",
    # Links and media
    "a",
    "img",
    "figure",
    "figcaption",
    # Tables
    "table",
    "thead",
    "tbody",
    "tfoot",
    "tr",
    "th",
    "td",
    # Code
    "pre",
    "code",
    # Quotes
    "blockquote",
    "q",
    "cite",
    # Semantic
    "div",
    "span",
    "section",
    "article",
}

ALLOWED_ATTRIBUTES = {
    "*": ["class", "id", "style"],  # Allowed on all tags
    "a": ["href", "target", "rel", "title"],
    "img": ["src", "alt", "width", "height", "loading"],
    "th": ["colspan", "rowspan", "scope"],
    "td": ["colspan", "rowspan"],
    "code": ["class"],  # For syntax highlighting
    "pre": ["class"],
}

# Dangerous URL schemes
DANGEROUS_SCHEMES = {"javascript", "vbscript", "data"}

# Allowed CSS properties (limited for security)
ALLOWED_STYLES = {
    "color",
    "background-color",
    "text-align",
    "font-weight",
    "font-style",
    "text-decoration",
    "margin",
    "padding",
    "border",
    "width",
    "height",
    "max-width",
    "max-height",
    "display",
    "float",
    "clear",
}


class SanitizerService:
    """Service for sanitizing HTML content to prevent XSS attacks."""

    def __init__(
        self,
        allowed_tags: set[str] = None,
        allowed_attributes: dict[str, list[str]] = None,
        strip_comments: bool = True,
    ):
        self.allowed_tags = allowed_tags or ALLOWED_TAGS
        self.allowed_attributes = allowed_attributes or ALLOWED_ATTRIBUTES
        self.strip_comments = strip_comments

    def sanitize(self, html_content: str) -> str:
        """
        Sanitize HTML content by removing dangerous elements.

        Args:
            html_content: Raw HTML string

        Returns:
            Sanitized HTML string
        """
        if not html_content:
            return ""

        # Remove HTML comments
        if self.strip_comments:
            html_content = re.sub(r"<!--.*?-->", "", html_content, flags=re.DOTALL)

        # Remove script tags completely
        html_content = re.sub(
            r"<script[^>]*>.*?</script>",
            "",
            html_content,
            flags=re.DOTALL | re.IGNORECASE,
        )

        # Remove style tags (inline styles are handled separately)
        html_content = re.sub(
            r"<style[^>]*>.*?</style>",
            "",
            html_content,
            flags=re.DOTALL | re.IGNORECASE,
        )

        # Remove on* event handlers
        html_content = re.sub(
            r'\s+on\w+\s*=\s*["\'][^"\']*["\']',
            "",
            html_content,
            flags=re.IGNORECASE,
        )
        html_content = re.sub(
            r"\s+on\w+\s*=\s*\S+",
            "",
            html_content,
            flags=re.IGNORECASE,
        )

        # Sanitize href/src attributes with dangerous protocols
        html_content = self._sanitize_urls(html_content)

        # Sanitize style attributes
        html_content = self._sanitize_styles(html_content)

        return html_content

    def _sanitize_urls(self, content: str) -> str:
        """Remove dangerous URL schemes from href/src attributes."""

        def replace_url(match: re.Match) -> str:
            attr = match.group(1)
            quote = match.group(2)
            url = match.group(3)

            # Check for dangerous schemes
            url_lower = url.lower().strip()
            for scheme in DANGEROUS_SCHEMES:
                if url_lower.startswith(f"{scheme}:"):
                    return f"{attr}={quote}#{quote}"

            return match.group(0)

        # Match href="..." or src="..."
        pattern = r'(href|src)\s*=\s*(["\'])([^"\']*)\2'
        return re.sub(pattern, replace_url, content, flags=re.IGNORECASE)

    def _sanitize_styles(self, content: str) -> str:
        """Sanitize inline style attributes."""

        def replace_style(match: re.Match) -> str:
            quote = match.group(1)
            style = match.group(2)

            # Parse and filter CSS properties
            safe_styles = []
            for declaration in style.split(";"):
                if ":" in declaration:
                    prop, value = declaration.split(":", 1)
                    prop = prop.strip().lower()
                    value = value.strip()

                    # Check if property is allowed
                    if prop in ALLOWED_STYLES:
                        # Remove url() and expression() from values
                        if "url(" not in value.lower() and "expression(" not in value.lower():
                            safe_styles.append(f"{prop}: {value}")

            if safe_styles:
                return f'style={quote}{"; ".join(safe_styles)}{quote}'
            return ""

        pattern = r'style\s*=\s*(["\'])([^"\']*)\1'
        return re.sub(pattern, replace_style, content, flags=re.IGNORECASE)

    def sanitize_text(self, text: str) -> str:
        """
        Escape HTML entities in plain text.

        Use this for text that should NOT contain any HTML.
        """
        if not text:
            return ""
        return html.escape(text)

    def sanitize_blocks(self, blocks: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """
        Sanitize Editor.js blocks.

        Args:
            blocks: List of Editor.js block objects

        Returns:
            Sanitized blocks
        """
        if not blocks:
            return []

        sanitized = []
        for block in blocks:
            block_type = block.get("type", "")
            data = block.get("data", {})

            if block_type == "paragraph":
                data["text"] = self.sanitize(data.get("text", ""))

            elif block_type == "header":
                data["text"] = self.sanitize(data.get("text", ""))

            elif block_type == "list":
                items = data.get("items", [])
                data["items"] = [self.sanitize(item) for item in items]

            elif block_type == "quote":
                data["text"] = self.sanitize(data.get("text", ""))
                data["caption"] = self.sanitize(data.get("caption", ""))

            elif block_type == "code":
                # Code blocks should be escaped, not sanitized
                data["code"] = self.sanitize_text(data.get("code", ""))

            elif block_type == "raw":
                # Raw HTML blocks need careful sanitization
                data["html"] = self.sanitize(data.get("html", ""))

            elif block_type == "table":
                content = data.get("content", [])
                data["content"] = [[self.sanitize(cell) for cell in row] for row in content]

            elif block_type == "image":
                # Validate image URL
                url = data.get("url", "")
                if url:
                    url_lower = url.lower().strip()
                    for scheme in DANGEROUS_SCHEMES:
                        if url_lower.startswith(f"{scheme}:"):
                            data["url"] = ""
                            break

                data["caption"] = self.sanitize(data.get("caption", ""))

            elif block_type == "embed":
                # Only allow trusted embed sources
                data["caption"] = self.sanitize(data.get("caption", ""))

            sanitized.append(
                {
                    "type": block_type,
                    "data": data,
                }
            )

        return sanitized


# Singleton instance
sanitizer_service = SanitizerService()



"""Export Service - Content export for CMS migration.

Exports content in various formats for migration to other systems.
"""

import csv
import io
import json
import zipfile
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Literal

from sqlalchemy import and_, select
from sqlalchemy.ext.asyncio import AsyncSession

from ..models import Entity, EntityValue, Relation
from ..utils import utcnow

ExportFormat = Literal["json", "csv", "wordpress", "markdown"]


@dataclass
class ExportOptions:
    """Export configuration options."""

    include_drafts: bool = False
    include_deleted: bool = False
    include_media: bool = True
    include_relations: bool = True
    include_metadata: bool = True
    date_from: datetime | None = None
    date_to: datetime | None = None


@dataclass
class ExportResult:
    """Export operation result."""

    format: str
    entity_count: int
    media_count: int
    file_size: int
    filename: str
    created_at: datetime


class ExportService:
    """
    Service for exporting content to various formats.

    Usage:
        export = ExportService(db)

        # Export to JSON
        data = await export.export_json("post")

        # Export to WordPress WXR
        wxr = await export.export_wordpress()

        # Export full site
        zip_data = await export.export_full_site()
    """

    def __init__(self, db: AsyncSession):
        self.db = db

    async def export_json(
        self,
        type_name: str = None,
        options: ExportOptions = None,
    ) -> dict:
        """
        Export content to JSON format.

        Args:
            type_name: Optional content type filter
            options: Export options

        Returns:
            JSON-serializable dict
        """
        options = options or ExportOptions()

        entities = await self._get_entities(type_name, options)

        result = {
            "version": "1.0",
            "exported_at": utcnow().isoformat(),
            "generator": "Focomy CMS",
            "content_types": {},
        }

        for entity in entities:
            type_name = entity.type
            if type_name not in result["content_types"]:
                result["content_types"][type_name] = []

            entity_data = await self._serialize_entity(entity, options)
            result["content_types"][type_name].append(entity_data)

        if options.include_media:
            result["media"] = await self._export_media_list()

        if options.include_relations:
            result["relations"] = await self._export_relations()

        return result

    async def export_csv(
        self,
        type_name: str,
        options: ExportOptions = None,
    ) -> str:
        """
        Export content type to CSV format.

        Args:
            type_name: Content type to export
            options: Export options

        Returns:
            CSV string
        """
        options = options or ExportOptions()

        entities = await self._get_entities(type_name, options)

        if not entities:
            return ""

        # Collect all field names
        all_fields = set()
        entity_values = {}

        for entity in entities:
            values = await self._get_entity_values(entity.id)
            entity_values[entity.id] = values
            all_fields.update(values.keys())

        # Sort fields for consistent output
        fields = ["id", "created_at", "updated_at"] + sorted(all_fields - {"id"})

        # Write CSV
        output = io.StringIO()
        writer = csv.DictWriter(output, fieldnames=fields, extrasaction="ignore")
        writer.writeheader()

        for entity in entities:
            row = {
                "id": entity.id,
                "created_at": entity.created_at.isoformat() if entity.created_at else "",
                "updated_at": entity.updated_at.isoformat() if entity.updated_at else "",
                **entity_values.get(entity.id, {}),
            }
            writer.writerow(row)

        return output.getvalue()

    async def export_wordpress(
        self,
        options: ExportOptions = None,
    ) -> str:
        """
        Export to WordPress WXR format.

        Args:
            options: Export options

        Returns:
            WXR XML string
        """
        options = options or ExportOptions()

        # Get posts
        posts = await self._get_entities("post", options)
        pages = await self._get_entities("page", options)

        # Build WXR
        wxr = self._build_wxr_header()

        # Add posts
        for post in posts:
            values = await self._get_entity_values(post.id)
            wxr += self._build_wxr_item(post, values, "post")

        # Add pages
        for page in pages:
            values = await self._get_entity_values(page.id)
            wxr += self._build_wxr_item(page, values, "page")

        wxr += self._build_wxr_footer()

        return wxr

    def _build_wxr_header(self) -> str:
        """Build WXR XML header."""
        return f"""<?xml version="1.0" encoding="UTF-8" ?>
<rss version="2.0"
    xmlns:excerpt="http://wordpress.org/export/1.2/excerpt/"
    xmlns:content="http://purl.org/rss/1.0/modules/content/"
    xmlns:wfw="http://wellformedweb.org/CommentAPI/"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:wp="http://wordpress.org/export/1.2/">
<channel>
    <title>Focomy Export</title>
    <link></link>
    <description>Exported from Focomy CMS</description>
    <pubDate>{utcnow().strftime('%a, %d %b %Y %H:%M:%S +0000')}</pubDate>
    <wp:wxr_version>1.2</wp:wxr_version>
"""

    def _build_wxr_item(self, entity: Entity, values: dict, post_type: str) -> str:
        """Build WXR item element."""
        title = values.get("title", "")
        content = self._extract_content_text(values.get("content", {}))
        slug = values.get("slug", "")
        status = "publish" if values.get("status") == "published" else "draft"
        date = entity.created_at.strftime("%Y-%m-%d %H:%M:%S") if entity.created_at else ""

        return f"""
    <item>
        <title><![CDATA[{title}]]></title>
        <link></link>
        <pubDate>{date}</pubDate>
        <dc:creator><![CDATA[admin]]></dc:creator>
        <content:encoded><![CDATA[{content}]]></content:encoded>
        <wp:post_id>{hash(entity.id) % 100000}</wp:post_id>
        <wp:post_date>{date}</wp:post_date>
        <wp:post_name>{slug}</wp:post_name>
        <wp:status>{status}</wp:status>
        <wp:post_type>{post_type}</wp:post_type>
    </item>
"""

    def _build_wxr_footer(self) -> str:
        """Build WXR XML footer."""
        return """
</channel>
</rss>
"""

    def _extract_content_text(self, content: Any) -> str:
        """Extract plain text from Editor.js content."""
        if not content:
            return ""

        if isinstance(content, str):
            return content

        if isinstance(content, dict):
            blocks = content.get("blocks", [])
            texts = []
            for block in blocks:
                block_type = block.get("type", "")
                data = block.get("data", {})

                if block_type == "paragraph":
                    texts.append(f"<p>{data.get('text', '')}</p>")
                elif block_type == "header":
                    level = data.get("level", 2)
                    texts.append(f"<h{level}>{data.get('text', '')}</h{level}>")
                elif block_type == "list":
                    items = data.get("items", [])
                    style = data.get("style", "unordered")
                    tag = "ol" if style == "ordered" else "ul"
                    li_items = "".join(f"<li>{item}</li>" for item in items)
                    texts.append(f"<{tag}>{li_items}</{tag}>")
                elif block_type == "quote":
                    texts.append(f"<blockquote>{data.get('text', '')}</blockquote>")
                elif block_type == "code":
                    texts.append(f"<pre><code>{data.get('code', '')}</code></pre>")

            return "\n".join(texts)

        return ""

    async def export_markdown(
        self,
        type_name: str = "post",
        options: ExportOptions = None,
    ) -> dict[str, str]:
        """
        Export content to Markdown files.

        Args:
            type_name: Content type to export
            options: Export options

        Returns:
            Dict of filename -> markdown content
        """
        options = options or ExportOptions()

        entities = await self._get_entities(type_name, options)

        files = {}

        for entity in entities:
            values = await self._get_entity_values(entity.id)
            slug = values.get("slug", entity.id)
            title = values.get("title", "Untitled")
            content = self._convert_to_markdown(values.get("content", {}))

            # Build frontmatter
            frontmatter = [
                "---",
                f'title: "{title}"',
                f"slug: {slug}",
                f"date: {entity.created_at.isoformat() if entity.created_at else ''}",
                f"status: {values.get('status', 'draft')}",
            ]

            if values.get("tags"):
                frontmatter.append(f"tags: {values['tags']}")

            frontmatter.append("---")
            frontmatter.append("")

            markdown = "\n".join(frontmatter) + content

            files[f"{slug}.md"] = markdown

        return files

    def _convert_to_markdown(self, content: Any) -> str:
        """Convert Editor.js content to Markdown."""
        if not content:
            return ""

        if isinstance(content, str):
            return content

        if isinstance(content, dict):
            blocks = content.get("blocks", [])
            texts = []

            for block in blocks:
                block_type = block.get("type", "")
                data = block.get("data", {})

                if block_type == "paragraph":
                    texts.append(data.get("text", ""))
                    texts.append("")
                elif block_type == "header":
                    level = data.get("level", 2)
                    texts.append("#" * level + " " + data.get("text", ""))
                    texts.append("")
                elif block_type == "list":
                    items = data.get("items", [])
                    style = data.get("style", "unordered")
                    for i, item in enumerate(items):
                        if style == "ordered":
                            texts.append(f"{i+1}. {item}")
                        else:
                            texts.append(f"- {item}")
                    texts.append("")
                elif block_type == "quote":
                    texts.append(f"> {data.get('text', '')}")
                    texts.append("")
                elif block_type == "code":
                    lang = data.get("language", "")
                    texts.append(f"```{lang}")
                    texts.append(data.get("code", ""))
                    texts.append("```")
                    texts.append("")
                elif block_type == "image":
                    url = data.get("file", {}).get("url", "")
                    caption = data.get("caption", "")
                    texts.append(f"![{caption}]({url})")
                    texts.append("")

            return "\n".join(texts)

        return ""

    async def export_full_site(
        self,
        options: ExportOptions = None,
    ) -> bytes:
        """
        Export entire site as a ZIP archive.

        Args:
            options: Export options

        Returns:
            ZIP file as bytes
        """
        options = options or ExportOptions()
        options.include_media = True
        options.include_relations = True

        zip_buffer = io.BytesIO()

        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zf:
            # Export JSON data
            json_data = await self.export_json(options=options)
            zf.writestr("data.json", json.dumps(json_data, indent=2, ensure_ascii=False))

            # Export content types definitions
            content_types = await self._export_content_type_definitions()
            zf.writestr("content_types.json", json.dumps(content_types, indent=2))

            # Export relations config
            relations_config = await self._export_relations_config()
            zf.writestr("relations.json", json.dumps(relations_config, indent=2))

            # Export media files
            if options.include_media:
                media_files = await self._get_media_files()
                for file_info in media_files:
                    try:
                        file_path = Path("uploads") / file_info["path"]
                        if file_path.exists():
                            zf.write(file_path, f"media/{file_info['path']}")
                    except Exception:
                        pass

        zip_buffer.seek(0)
        return zip_buffer.read()

    async def _get_entities(
        self,
        type_name: str = None,
        options: ExportOptions = None,
    ) -> list[Entity]:
        """Get entities for export."""
        options = options or ExportOptions()

        conditions = []

        if type_name:
            conditions.append(Entity.type == type_name)

        if not options.include_deleted:
            conditions.append(Entity.deleted_at.is_(None))

        if not options.include_drafts:
            # Would need to check status field
            pass

        if options.date_from:
            conditions.append(Entity.created_at >= options.date_from)

        if options.date_to:
            conditions.append(Entity.created_at <= options.date_to)

        query = select(Entity)
        if conditions:
            query = query.where(and_(*conditions))

        result = await self.db.execute(query)
        return result.scalars().all()

    async def _get_entity_values(self, entity_id: str) -> dict:
        """Get all values for an entity."""
        query = select(EntityValue).where(EntityValue.entity_id == entity_id)
        result = await self.db.execute(query)
        values = result.scalars().all()

        return {v.field_name: v.value_json if v.value_json else v.value_text for v in values}

    async def _serialize_entity(
        self,
        entity: Entity,
        options: ExportOptions,
    ) -> dict:
        """Serialize entity for export."""
        values = await self._get_entity_values(entity.id)

        data = {
            "id": entity.id,
            "type": entity.type,
            "values": values,
            "created_at": entity.created_at.isoformat() if entity.created_at else None,
            "updated_at": entity.updated_at.isoformat() if entity.updated_at else None,
        }

        if options.include_metadata:
            data["created_by"] = entity.created_by
            data["updated_by"] = entity.updated_by

        if options.include_relations:
            data["relations"] = await self._get_entity_relations(entity.id)

        return data

    async def _get_entity_relations(self, entity_id: str) -> list[dict]:
        """Get relations for an entity."""
        query = select(Relation).where(Relation.from_entity_id == entity_id)
        result = await self.db.execute(query)
        relations = result.scalars().all()

        return [
            {
                "type": r.relation_type,
                "to_entity_id": r.to_entity_id,
                "position": r.position,
            }
            for r in relations
        ]

    async def _export_media_list(self) -> list[dict]:
        """Export media entity list."""
        media_entities = await self._get_entities("media")
        return [
            await self._serialize_entity(e, ExportOptions(include_relations=False))
            for e in media_entities
        ]

    async def _export_relations(self) -> list[dict]:
        """Export all relations."""
        query = select(Relation)
        result = await self.db.execute(query)
        relations = result.scalars().all()

        return [
            {
                "from_entity_id": r.from_entity_id,
                "to_entity_id": r.to_entity_id,
                "relation_type": r.relation_type,
                "position": r.position,
            }
            for r in relations
        ]

    async def _export_content_type_definitions(self) -> dict:
        """Export content type YAML definitions."""
        # Would read from content_types directory
        return {}

    async def _export_relations_config(self) -> dict:
        """Export relations configuration."""
        # Would read from relations.yaml
        return {}

    async def _get_media_files(self) -> list[dict]:
        """Get media file information."""
        media_entities = await self._get_entities("media")
        files = []

        for entity in media_entities:
            values = await self._get_entity_values(entity.id)
            if values.get("path"):
                files.append(
                    {
                        "id": entity.id,
                        "path": values.get("path"),
                        "filename": values.get("filename"),
                    }
                )

        return files


def get_export_service(db: AsyncSession) -> ExportService:
    return ExportService(db)



"""Media model - uploaded files management."""

import uuid
from datetime import datetime

from sqlalchemy import DateTime, Integer, String, Text
from sqlalchemy.orm import Mapped, mapped_column

from ..database import Base


def generate_uuid() -> str:
    return str(uuid.uuid4())


class Media(Base):
    """
    Uploaded media files.

    Stores metadata about uploaded files.
    Actual files are stored in the filesystem.
    """

    __tablename__ = "media"

    id: Mapped[str] = mapped_column(
        String(36),
        primary_key=True,
        default=generate_uuid,
    )
    filename: Mapped[str] = mapped_column(
        String(255),
        nullable=False,
    )
    stored_path: Mapped[str] = mapped_column(
        String(500),
        nullable=False,
    )
    mime_type: Mapped[str] = mapped_column(
        String(100),
        nullable=False,
    )
    size: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
    )
    width: Mapped[int | None] = mapped_column(
        Integer,
        nullable=True,
    )
    height: Mapped[int | None] = mapped_column(
        Integer,
        nullable=True,
    )
    alt_text: Mapped[str | None] = mapped_column(
        Text,
        nullable=True,
    )
    file_hash: Mapped[str | None] = mapped_column(
        String(64),
        nullable=True,
        index=True,
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime,
        nullable=False,
        default=datetime.utcnow,
    )
    created_by: Mapped[str | None] = mapped_column(
        String(36),
        nullable=True,
    )
    updated_at: Mapped[datetime | None] = mapped_column(
        DateTime,
        nullable=True,
        onupdate=datetime.utcnow,
    )
    updated_by: Mapped[str | None] = mapped_column(
        String(36),
        nullable=True,
    )
    deleted_at: Mapped[datetime | None] = mapped_column(
        DateTime,
        nullable=True,
        index=True,
    )

    def __repr__(self) -> str:
        return f"<Media(id={self.id}, filename={self.filename})>"



"""Bulk Operations Service - Batch entity operations.

Provides efficient bulk create, update, and delete operations.
"""

from dataclasses import dataclass
from typing import Any

from sqlalchemy.ext.asyncio import AsyncSession

from .entity import EntityService


@dataclass
class BulkResult:
    """Result of bulk operation."""

    success_count: int
    error_count: int
    errors: list[dict]


class BulkOperationService:
    """
    Service for bulk entity operations.

    Usage:
        bulk = BulkOperationService(db)

        # Bulk update
        result = await bulk.update_many(
            entity_ids=["id1", "id2", "id3"],
            data={"status": "published"},
            user_id="user123",
        )

        # Bulk delete
        result = await bulk.delete_many(
            entity_ids=["id1", "id2"],
            user_id="user123",
        )
    """

    def __init__(self, db: AsyncSession):
        self.db = db
        self.entity_svc = EntityService(db)

    async def update_many(
        self,
        entity_ids: list[str],
        data: dict[str, Any],
        user_id: str | None = None,
        validate: bool = True,
    ) -> BulkResult:
        """
        Update multiple entities with the same data.

        Args:
            entity_ids: List of entity IDs to update
            data: Fields to update (same for all entities)
            user_id: User performing the update
            validate: Whether to validate data against content type

        Returns:
            BulkResult with success/error counts
        """
        success_count = 0
        errors = []

        for entity_id in entity_ids:
            try:
                await self.entity_svc.update(
                    entity_id,
                    data,
                    user_id=user_id,
                    validate=validate,
                )
                success_count += 1
            except Exception as e:
                errors.append(
                    {
                        "entity_id": entity_id,
                        "error": str(e),
                    }
                )

        return BulkResult(
            success_count=success_count,
            error_count=len(errors),
            errors=errors,
        )

    async def delete_many(
        self,
        entity_ids: list[str],
        user_id: str | None = None,
        hard: bool = False,
    ) -> BulkResult:
        """
        Delete multiple entities.

        Args:
            entity_ids: List of entity IDs to delete
            user_id: User performing the delete
            hard: If True, permanently delete (not recommended)

        Returns:
            BulkResult with success/error counts
        """
        success_count = 0
        errors = []

        for entity_id in entity_ids:
            try:
                result = await self.entity_svc.delete(
                    entity_id,
                    user_id=user_id,
                    hard=hard,
                )
                if result:
                    success_count += 1
                else:
                    errors.append(
                        {
                            "entity_id": entity_id,
                            "error": "Entity not found",
                        }
                    )
            except Exception as e:
                errors.append(
                    {
                        "entity_id": entity_id,
                        "error": str(e),
                    }
                )

        return BulkResult(
            success_count=success_count,
            error_count=len(errors),
            errors=errors,
        )

    async def restore_many(
        self,
        entity_ids: list[str],
        user_id: str | None = None,
    ) -> BulkResult:
        """
        Restore multiple soft-deleted entities.

        Args:
            entity_ids: List of entity IDs to restore
            user_id: User performing the restore

        Returns:
            BulkResult with success/error counts
        """
        success_count = 0
        errors = []

        for entity_id in entity_ids:
            try:
                result = await self.entity_svc.restore(
                    entity_id,
                    user_id=user_id,
                )
                if result:
                    success_count += 1
                else:
                    errors.append(
                        {
                            "entity_id": entity_id,
                            "error": "Entity not found or not deleted",
                        }
                    )
            except Exception as e:
                errors.append(
                    {
                        "entity_id": entity_id,
                        "error": str(e),
                    }
                )

        return BulkResult(
            success_count=success_count,
            error_count=len(errors),
            errors=errors,
        )

    async def create_many(
        self,
        entity_type: str,
        items: list[dict[str, Any]],
        user_id: str | None = None,
    ) -> BulkResult:
        """
        Create multiple entities.

        Args:
            entity_type: Content type name
            items: List of entity data dicts
            user_id: User creating the entities

        Returns:
            BulkResult with success/error counts
        """
        success_count = 0
        errors = []

        for i, data in enumerate(items):
            try:
                await self.entity_svc.create(
                    entity_type,
                    data,
                    user_id=user_id,
                )
                success_count += 1
            except Exception as e:
                errors.append(
                    {
                        "index": i,
                        "data": data,
                        "error": str(e),
                    }
                )

        return BulkResult(
            success_count=success_count,
            error_count=len(errors),
            errors=errors,
        )

    async def update_status(
        self,
        entity_ids: list[str],
        status: str,
        user_id: str | None = None,
    ) -> BulkResult:
        """
        Update status for multiple entities.

        Convenience method for common bulk status change.

        Args:
            entity_ids: List of entity IDs
            status: New status value
            user_id: User performing the update

        Returns:
            BulkResult with success/error counts
        """
        return await self.update_many(
            entity_ids=entity_ids,
            data={"status": status},
            user_id=user_id,
        )

    async def assign_to(
        self,
        entity_ids: list[str],
        user_id: str,
        assigned_by: str | None = None,
    ) -> BulkResult:
        """
        Assign multiple entities to a user.

        Args:
            entity_ids: List of entity IDs
            user_id: User to assign to
            assigned_by: User performing the assignment

        Returns:
            BulkResult with success/error counts
        """
        return await self.update_many(
            entity_ids=entity_ids,
            data={"assigned_to": user_id},
            user_id=assigned_by,
        )


def get_bulk_service(db: AsyncSession) -> BulkOperationService:
    return BulkOperationService(db)



"""Admin theme routes - theme management and customization."""

from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import HTMLResponse, RedirectResponse
from sqlalchemy.ext.asyncio import AsyncSession

from ..database import get_db
from ..models import Entity
from ..services.entity import EntityService
from .helpers import templates, require_admin, require_admin_api, get_context

router = APIRouter()


# === Theme Management ===


@router.get("/themes", response_class=HTMLResponse)
async def themes_page(
    request: Request,
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Theme management page."""
    from ..services.settings import SettingsService
    from ..services.theme import theme_service

    themes_data = []
    for _name, theme in theme_service.get_all_themes().items():
        themes_data.append(
            {
                "name": theme.name,
                "label": theme.label,
                "description": theme.description,
                "version": theme.version,
                "author": theme.author,
                "preview": getattr(theme, "preview", None),
            }
        )

    # Get active theme from database settings
    settings_svc = SettingsService(db)
    theme_settings = await settings_svc.get_by_category("theme")
    active_theme = theme_settings.get("active", "default")

    context = await get_context(request, db, current_user, "themes")
    context.update(
        {
            "themes": themes_data,
            "active_theme": active_theme,
            "message": request.query_params.get("message"),
        }
    )

    return templates.TemplateResponse("admin/themes.html", context)


@router.post("/themes/{theme_name}/activate", response_class=HTMLResponse)
async def activate_theme(
    request: Request,
    theme_name: str,
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Activate a theme."""
    from ..services.settings import SettingsService
    from ..services.theme import theme_service

    # Validate theme exists
    themes = theme_service.get_all_themes()
    if theme_name not in themes:
        raise HTTPException(status_code=404, detail="Theme not found")

    # Save to settings
    settings_svc = SettingsService(db)
    entity_svc = EntityService(db)
    user_data = entity_svc.serialize(current_user)
    await settings_svc.set(
        "theme.active", theme_name, category="theme", user_id=user_data.get("id")
    )

    # Update theme service
    theme_service.set_current_theme(theme_name)

    return RedirectResponse(url="/admin/themes?message=テーマを変更しました", status_code=303)


# === Theme Customization ===


@router.get("/themes/{theme_name}/customize", response_class=HTMLResponse)
async def customize_theme_page(
    request: Request,
    theme_name: str,
    db: AsyncSession = Depends(get_db),
    current_user: Entity = Depends(require_admin),
):
    """Theme customization page."""
    from ..services.theme import theme_service

    # Validate theme exists
    theme = theme_service.get_theme(theme_name)
    if not theme:
        raise HTTPException(status_code=404, detail="Theme not found")

    # Get customizable settings
    settings = theme_service.get_customizable_settings(theme_name)

    # Group settings by category
    grouped_settings = {}
    for setting in settings:
        category = setting.get("category", "other")
        if category not in grouped_settings:
            grouped_settings[category] = []
        grouped_settings[category].append(setting)

    # Debug log
    import structlog
    logger = structlog.get_logger()
    logger.info(
        "customize_theme_page",
        theme_name=theme_name,
        settings_count=len(settings),
        grouped_keys=list(grouped_settings.keys()),
        colors_count=len(grouped_settings.get("colors", [])),
    )

    context = await get_context(request, db, current_user, "themes")
    context.update({
        "theme": {
            "name": theme.name,
            "label": theme.label,
        },
        "settings": settings,
        "grouped_settings": grouped_settings,
        "message": request.query_params.get("message"),
    })

    return templates.TemplateResponse("admin/customize.html", context)


@router.get("/api/theme/settings")
async def get_theme_settings(
    request: Request,
    theme_name: str = None,
    current_user: Entity = Depends(require_admin_api),
):
    """Get theme customization settings."""
    from ..services.settings import SettingsService
    from ..services.theme import theme_service

    # Get active theme if not specified
    if not theme_name:
        db = request.state.db
        settings_svc = SettingsService(db)
        theme_settings = await settings_svc.get_by_category("theme")
        theme_name = theme_settings.get("active", "default")

    theme = theme_service.get_theme(theme_name)
    if not theme:
        raise HTTPException(status_code=404, detail="Theme not found")

    settings = theme_service.get_customizable_settings(theme_name)
    customizations = theme_service.get_customizations(theme_name)

    return {
        "theme_name": theme_name,
        "settings": settings,
        "customizations": customizations,
    }


@router.post("/api/theme/settings")
async def save_theme_settings(
    request: Request,
    current_user: Entity = Depends(require_admin_api),
):
    """Save theme customization settings."""
    from ..services.theme import theme_service

    try:
        body = await request.json()
    except Exception:
        raise HTTPException(status_code=400, detail="Invalid JSON")

    theme_name = body.get("theme_name")
    values = body.get("values", {})

    if not theme_name:
        raise HTTPException(status_code=400, detail="theme_name is required")

    theme = theme_service.get_theme(theme_name)
    if not theme:
        raise HTTPException(status_code=404, detail="Theme not found")

    success = theme_service.set_customizations(values, theme_name)

    if success:
        return {"success": True, "message": "カスタマイズを保存しました"}
    else:
        raise HTTPException(status_code=500, detail="Failed to save customizations")


@router.post("/api/theme/preview-css")
async def preview_theme_css(
    request: Request,
    current_user: Entity = Depends(require_admin_api),
):
    """Generate preview CSS with temporary values."""
    from fastapi.responses import PlainTextResponse

    from ..services.theme import theme_service

    try:
        body = await request.json()
    except Exception:
        raise HTTPException(status_code=400, detail="Invalid JSON")

    theme_name = body.get("theme_name")
    preview_values = body.get("values", {})

    theme = theme_service.get_theme(theme_name)
    if not theme:
        raise HTTPException(status_code=404, detail="Theme not found")

    css = theme_service.generate_preview_css(preview_values, theme_name)

    return PlainTextResponse(content=css, media_type="text/css")



"""WidgetService - sidebar and footer widget management.

Provides customizable widgets for sidebars and footers.
Each widget type has its own rendering logic.
"""

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from sqlalchemy.ext.asyncio import AsyncSession

from .entity import EntityService


@dataclass
class WidgetContext:
    """Context passed to widget renderers."""

    db: AsyncSession
    entity_svc: EntityService
    request_context: dict[str, Any]


class BaseWidget(ABC):
    """Base class for all widgets."""

    name: str = "base"
    label: str = "Base Widget"

    @abstractmethod
    async def render(self, config: dict, context: WidgetContext) -> str:
        """Render the widget HTML."""
        pass


class RecentPostsWidget(BaseWidget):
    """Display recent posts."""

    name = "recent_posts"
    label = "Recent Posts"

    async def render(self, config: dict, context: WidgetContext) -> str:
        limit = config.get("limit", 5)

        posts = await context.entity_svc.find(
            "post",
            limit=limit,
            order_by="-created_at",
            filters={"status": "published"},
        )

        items = []
        for p in posts:
            data = context.entity_svc.serialize(p)
            title = data.get("title", "Untitled")
            slug = data.get("slug", "")
            date = data.get("created_at", "")[:10] if data.get("created_at") else ""
            items.append(
                f"""
                <li class="widget-post-item">
                    <a href="/post/{slug}">{title}</a>
                    <span class="widget-post-date">{date}</span>
                </li>
            """
            )

        return f"""
            <ul class="widget-posts-list">
                {"".join(items)}
            </ul>
        """


class CategoriesWidget(BaseWidget):
    """Display category list."""

    name = "categories"
    label = "Categories"

    async def render(self, config: dict, context: WidgetContext) -> str:
        config.get("show_count", True)

        categories = await context.entity_svc.find(
            "category",
            limit=50,
            order_by="name",
        )

        items = []
        for cat in categories:
            data = context.entity_svc.serialize(cat)
            name = data.get("name", "")
            slug = data.get("slug", "")
            items.append(
                f"""
                <li class="widget-category-item">
                    <a href="/category/{slug}">{name}</a>
                </li>
            """
            )

        return f"""
            <ul class="widget-categories-list">
                {"".join(items)}
            </ul>
        """


class ChannelListWidget(BaseWidget):
    """Display channel list."""

    name = "channel_list"
    label = "Channel List"

    async def render(self, config: dict, context: WidgetContext) -> str:
        channels = await context.entity_svc.find(
            "channel",
            limit=50,
            order_by="title",
        )

        items = []
        for ch in channels:
            data = context.entity_svc.serialize(ch)
            title = data.get("title", "")
            slug = data.get("slug", "")
            items.append(
                f"""
                <li class="widget-channel-item">
                    <a href="/channel/{slug}">{title}</a>
                </li>
            """
            )

        return f"""
            <ul class="widget-channels-list">
                {"".join(items)}
            </ul>
        """


class SeriesListWidget(BaseWidget):
    """Display series list."""

    name = "series_list"
    label = "Series List"

    async def render(self, config: dict, context: WidgetContext) -> str:
        series_list = await context.entity_svc.find(
            "series",
            limit=50,
            order_by="title",
        )

        items = []
        for s in series_list:
            data = context.entity_svc.serialize(s)
            title = data.get("title", "")
            slug = data.get("slug", "")
            items.append(
                f"""
                <li class="widget-series-item">
                    <a href="/series/{slug}">{title}</a>
                </li>
            """
            )

        return f"""
            <ul class="widget-series-list">
                {"".join(items)}
            </ul>
        """


class SearchWidget(BaseWidget):
    """Display search form."""

    name = "search"
    label = "Search"

    async def render(self, config: dict, context: WidgetContext) -> str:
        placeholder = config.get("placeholder", "Search...")
        return f"""
            <form action="/search" method="GET" class="widget-search-form">
                <input type="text" name="q" placeholder="{placeholder}" class="widget-search-input">
                <button type="submit" class="widget-search-button">Search</button>
            </form>
        """


class ArchivesWidget(BaseWidget):
    """Display monthly archives."""

    name = "archives"
    label = "Archives"

    async def render(self, config: dict, context: WidgetContext) -> str:
        limit = config.get("limit", 12)

        # Get posts to extract unique months
        posts = await context.entity_svc.find(
            "post",
            limit=500,
            order_by="-created_at",
            filters={"status": "published"},
        )

        # Extract unique year-month combinations
        months = {}
        for p in posts:
            data = context.entity_svc.serialize(p)
            created = data.get("created_at", "")
            if created and len(created) >= 7:
                ym = created[:7]
                months[ym] = months.get(ym, 0) + 1

        # Sort and limit
        sorted_months = sorted(months.items(), reverse=True)[:limit]

        items = []
        for ym, count in sorted_months:
            year, month = ym.split("-")
            label = f"{year}年{month}月"
            items.append(
                f"""
                <li class="widget-archive-item">
                    <a href="/archive/{year}/{month}">{label}</a>
                    <span class="widget-archive-count">({count})</span>
                </li>
            """
            )

        return f"""
            <ul class="widget-archives-list">
                {"".join(items)}
            </ul>
        """


class CustomHtmlWidget(BaseWidget):
    """Display custom HTML content."""

    name = "custom_html"
    label = "Custom HTML"

    async def render(self, config: dict, context: WidgetContext) -> str:
        html = config.get("html", "")
        return f'<div class="widget-custom-html">{html}</div>'


class WidgetService:
    """
    Widget management and rendering service.

    Handles widget CRUD and rendering for all areas.
    """

    AREAS = ("sidebar", "footer_1", "footer_2", "footer_3")

    # Register widget types
    WIDGET_TYPES: dict[str, type[BaseWidget]] = {
        "recent_posts": RecentPostsWidget,
        "categories": CategoriesWidget,
        "channel_list": ChannelListWidget,
        "series_list": SeriesListWidget,
        "search": SearchWidget,
        "archives": ArchivesWidget,
        "custom_html": CustomHtmlWidget,
    }

    def __init__(self, db: AsyncSession):
        self.db = db
        self.entity_svc = EntityService(db)

    async def get_widgets_for_area(self, area: str) -> list[dict[str, Any]]:
        """Get all active widgets for an area."""
        if area not in self.AREAS:
            return []

        entities = await self.entity_svc.find(
            "widget",
            limit=50,
            order_by="sort_order",
            filters={"area": area, "is_active": True},
        )

        widgets = []
        for e in entities:
            data = self.entity_svc.serialize(e)
            # Parse config JSON
            config = data.get("config")
            if isinstance(config, str):
                try:
                    config = json.loads(config)
                except json.JSONDecodeError:
                    config = {}
            data["config"] = config or {}
            widgets.append(data)

        return widgets

    async def render_area(
        self,
        area: str,
        request_context: dict[str, Any] = None,
    ) -> str:
        """Render all widgets for an area."""
        widgets = await self.get_widgets_for_area(area)
        if not widgets:
            return ""

        context = WidgetContext(
            db=self.db,
            entity_svc=self.entity_svc,
            request_context=request_context or {},
        )

        rendered = []
        for widget_data in widgets:
            widget_type = widget_data.get("widget_type", "custom_html")
            title = widget_data.get("title", "")
            config = widget_data.get("config", {})

            # For custom_html, pass the HTML in config
            if widget_type == "custom_html":
                config["html"] = widget_data.get("custom_html", "")

            # Get widget class
            widget_class = self.WIDGET_TYPES.get(widget_type)
            if not widget_class:
                continue

            widget = widget_class()

            try:
                content = await widget.render(config, context)
            except Exception as e:
                content = f'<p class="widget-error">Widget error: {str(e)}</p>'

            title_html = f'<h3 class="widget-title">{title}</h3>' if title else ""
            rendered.append(
                f"""
                <div class="widget widget--{widget_type}">
                    {title_html}
                    <div class="widget-content">
                        {content}
                    </div>
                </div>
            """
            )

        return "\n".join(rendered)

    async def render_all_areas(
        self,
        request_context: dict[str, Any] = None,
    ) -> dict[str, str]:
        """Render all widget areas."""
        return {area: await self.render_area(area, request_context) for area in self.AREAS}

    async def create_widget(
        self,
        widget_type: str,
        area: str,
        title: str = "",
        config: dict = None,
        custom_html: str = "",
        sort_order: int = 0,
        user_id: str = None,
    ):
        """Create a new widget."""
        data = {
            "title": title,
            "widget_type": widget_type,
            "area": area,
            "sort_order": sort_order,
            "config": json.dumps(config or {}),
            "custom_html": custom_html,
            "is_active": True,
        }
        return await self.entity_svc.create("widget", data, user_id)

    async def update_widget(
        self,
        widget_id: str,
        data: dict[str, Any],
        user_id: str = None,
    ):
        """Update a widget."""
        if "config" in data and isinstance(data["config"], dict):
            data["config"] = json.dumps(data["config"])
        return await self.entity_svc.update(widget_id, data, user_id, create_revision=False)

    async def delete_widget(self, widget_id: str, user_id: str = None) -> bool:
        """Delete a widget."""
        return await self.entity_svc.delete(widget_id, user_id)

    async def reorder_widgets(
        self,
        area: str,
        widget_orders: list[dict[str, Any]],
        user_id: str = None,
    ) -> bool:
        """Reorder widgets in an area."""
        for order in widget_orders:
            widget_id = order["id"]
            sort_order = order.get("sort_order", 0)
            await self.update_widget(
                widget_id,
                {"sort_order": sort_order},
                user_id=user_id,
            )
        return True

    @classmethod
    def get_available_widget_types(cls) -> list[dict[str, str]]:
        """Get list of available widget types."""
        return [
            {"value": name, "label": widget_class.label}
            for name, widget_class in cls.WIDGET_TYPES.items()
        ]



"""Block Converter - HTML/Gutenberg to Editor.js blocks."""

import json
import re

from bs4 import BeautifulSoup


class BlockConverter:
    """Convert WordPress HTML/Gutenberg to Editor.js format."""

    MAX_DEPTH = 10
    MAX_SIZE = 5_000_000  # 5MB

    # Gutenbergブロックパターン
    GUTENBERG_PATTERN = re.compile(
        r"<!-- wp:(\S+?)(?:\s+(\{.*?\}))?\s*-->(.*?)<!-- /wp:\1\s*-->",
        re.DOTALL,
    )

    # ショートコードパターン
    # [shortcode attr="value"] or [shortcode attr="value"]content[/shortcode]
    # Note: ショートコード名はハイフン許可（contact-form等）
    SHORTCODE_PATTERN = re.compile(
        r'\[([\w-]+)((?:\s+[\w-]+=["\'][^"\']*["\']|\s+[\w-]+=\S+)*)\s*/?]'
        r'(?:(.*?)\[/\1])?',
        re.DOTALL | re.IGNORECASE,
    )

    def convert(self, html: str) -> dict:
        """Convert HTML to Editor.js block format.

        Args:
            html: WordPress HTML content (Gutenberg or classic)

        Returns:
            Editor.js format dict with "blocks" list
        """
        if not html or not html.strip():
            return {"blocks": []}

        # サイズチェック
        if len(html) > self.MAX_SIZE:
            return {
                "blocks": [
                    {
                        "type": "raw",
                        "data": {"html": html[:1000] + "... (content truncated)"},
                    }
                ]
            }

        # Gutenbergブロック検出
        if "<!-- wp:" in html:
            blocks = self._parse_gutenberg(html)
        else:
            blocks = self._parse_html(html)

        return {"blocks": blocks}

    def _parse_gutenberg(self, html: str) -> list[dict]:
        """Parse Gutenberg block comments."""
        blocks = []

        for match in self.GUTENBERG_PATTERN.finditer(html):
            block_name = match.group(1)
            attrs_str = match.group(2)
            content = match.group(3)

            # 属性パース
            try:
                attrs = json.loads(attrs_str) if attrs_str else {}
            except json.JSONDecodeError:
                attrs = {}

            # ブロック変換
            block = self._convert_block(block_name, attrs, content)
            if block:
                if isinstance(block, list):
                    blocks.extend(block)
                else:
                    blocks.append(block)

        # マッチしなかった場合は全体をrawで保持
        if not blocks and html.strip():
            blocks = [{"type": "raw", "data": {"html": html.strip()}}]

        return blocks

    def _convert_block(
        self, name: str, attrs: dict, content: str, depth: int = 0
    ) -> dict | None:
        """Convert a single Gutenberg block.

        Args:
            name: Block name (e.g., "paragraph", "core/heading")
            attrs: Block attributes from JSON
            content: Inner HTML content
            depth: Nesting depth for recursion limit

        Returns:
            Editor.js block dict or None if empty
        """
        if depth > self.MAX_DEPTH:
            return {"type": "raw", "data": {"html": content.strip()}}

        # core/ プレフィックス除去
        name = name.replace("core/", "")

        if name == "paragraph":
            return self._convert_paragraph(content, attrs)
        elif name == "heading":
            return self._convert_header(content, attrs)
        elif name == "list":
            return self._convert_list(content, attrs)
        elif name == "image":
            return self._convert_image(content, attrs)
        elif name == "quote":
            return self._convert_quote(content, attrs)
        elif name == "code" or name == "preformatted":
            return self._convert_code(content, attrs)
        elif name == "separator":
            return self._convert_delimiter()
        elif name == "table":
            return self._convert_table(content, attrs)
        elif name == "embed" or name.startswith("core-embed/"):
            return self._convert_embed(content, attrs, name)
        elif name == "html":
            return self._convert_raw(content)
        elif name == "freeform":
            # クラシックブロック - 中身をパースして複数ブロックに分解
            return self._convert_freeform(content, depth)
        elif name in ("columns", "column", "group", "cover"):
            # ネストブロック - 中のブロックを再帰的にパース
            return self._convert_nested(content, depth)

        # 未対応ブロック → raw
        return {"type": "raw", "data": {"html": content.strip()}} if content.strip() else None

    def _convert_paragraph(self, content: str, attrs: dict) -> dict | None:
        """Convert paragraph block.

        Args:
            content: HTML content inside the block
            attrs: Block attributes (e.g., align)

        Returns:
            Editor.js paragraph block or None if empty
        """
        soup = BeautifulSoup(content, "html.parser")
        p = soup.find("p")

        if not p:
            text = content.strip()
        else:
            # decode_contents() preserves inner HTML (inline tags)
            text = str(p.decode_contents())

        if not text:
            return None  # 空paragraphはスキップ

        block = {"type": "paragraph", "data": {"text": text}}

        # 配置属性
        align = attrs.get("align")
        if align in ("center", "right"):
            block["tunes"] = {"alignmentTune": {"alignment": align}}

        return block

    def _convert_header(self, content: str, attrs: dict) -> dict | None:
        """Convert heading block."""
        soup = BeautifulSoup(content, "html.parser")

        # h1-h6を探す
        heading = None
        for level in range(1, 7):
            heading = soup.find(f"h{level}")
            if heading:
                break

        if heading:
            text = heading.get_text(strip=True)
            detected_level = int(heading.name[1])
        else:
            text = content.strip()
            detected_level = attrs.get("level", 2)

        if not text:
            return None

        level = attrs.get("level", detected_level)
        return {"type": "header", "data": {"text": text, "level": level}}

    def _convert_list(self, content: str, attrs: dict) -> dict | None:
        """Convert list block."""
        soup = BeautifulSoup(content, "html.parser")

        # ul or ol
        ul = soup.find("ul")
        ol = soup.find("ol")
        list_el = ol if ol else ul

        if not list_el:
            return None

        style = "ordered" if ol else "unordered"
        items = []
        for li in list_el.find_all("li", recursive=False):
            items.append(str(li.decode_contents()))

        if not items:
            return None

        return {"type": "list", "data": {"style": style, "items": items}}

    def _convert_image(self, content: str, attrs: dict) -> dict | None:
        """Convert image block."""
        soup = BeautifulSoup(content, "html.parser")
        img = soup.find("img")

        if not img:
            return None

        url = img.get("src", "")
        alt = img.get("alt", "")

        # figcaptionからキャプション取得
        caption = ""
        figcaption = soup.find("figcaption")
        if figcaption:
            caption = figcaption.get_text(strip=True)

        if not url:
            return None

        return {
            "type": "image",
            "data": {
                "file": {"url": url},
                "caption": caption or alt,
            },
        }

    def _convert_quote(self, content: str, attrs: dict) -> dict | None:
        """Convert quote block."""
        soup = BeautifulSoup(content, "html.parser")
        blockquote = soup.find("blockquote")

        if blockquote:
            # 引用テキスト
            p = blockquote.find("p")
            text = str(p.decode_contents()) if p else blockquote.get_text(strip=True)

            # 引用元（cite）
            cite = blockquote.find("cite")
            caption = cite.get_text(strip=True) if cite else ""
        else:
            text = content.strip()
            caption = ""

        if not text:
            return None

        return {"type": "quote", "data": {"text": text, "caption": caption}}

    def _convert_code(self, content: str, attrs: dict) -> dict | None:
        """Convert code block."""
        soup = BeautifulSoup(content, "html.parser")

        # pre > code or just pre or just code
        code = soup.find("code")
        pre = soup.find("pre")

        if code:
            code_text = code.get_text()
        elif pre:
            code_text = pre.get_text()
        else:
            code_text = content.strip()

        if not code_text:
            return None

        return {"type": "code", "data": {"code": code_text}}

    def _convert_delimiter(self) -> dict:
        """Convert separator/delimiter block."""
        return {"type": "delimiter", "data": {}}

    def _convert_table(self, content: str, attrs: dict) -> dict | None:
        """Convert table block."""
        soup = BeautifulSoup(content, "html.parser")
        table = soup.find("table")

        if not table:
            return None

        rows = []
        for tr in table.find_all("tr"):
            cells = []
            for cell in tr.find_all(["td", "th"]):
                cells.append(str(cell.decode_contents()))
            if cells:
                rows.append(cells)

        if not rows:
            return None

        return {"type": "table", "data": {"content": rows}}

    def _convert_embed(self, content: str, attrs: dict, name: str) -> dict | None:
        """Convert embed block (YouTube, Vimeo, etc.)."""
        url = attrs.get("url", "")

        # サービス判定
        if "core-embed/" in name:
            service = name.split("/")[-1]
        else:
            service = attrs.get("providerNameSlug", "")

        # URLからサービスを推測
        if not service:
            if "youtube" in url or "youtu.be" in url:
                service = "youtube"
            elif "vimeo" in url:
                service = "vimeo"
            elif "twitter" in url or "x.com" in url:
                service = "twitter"
            elif "spotify" in url:
                service = "spotify"

        # 埋め込みURL生成
        embed_url = url
        if service == "youtube":
            # YouTube URL → embed URL変換
            if "watch?v=" in url:
                video_id = url.split("watch?v=")[-1].split("&")[0]
                embed_url = f"https://www.youtube.com/embed/{video_id}"
            elif "youtu.be/" in url:
                video_id = url.split("youtu.be/")[-1].split("?")[0]
                embed_url = f"https://www.youtube.com/embed/{video_id}"

        if not url:
            return None

        return {
            "type": "embed",
            "data": {
                "service": service,
                "source": url,
                "embed": embed_url,
                "caption": attrs.get("caption", ""),
            },
        }

    def _convert_raw(self, content: str) -> dict | None:
        """Convert HTML block (raw HTML)."""
        if not content.strip():
            return None
        return {"type": "raw", "data": {"html": content.strip()}}

    def _convert_freeform(self, content: str, depth: int) -> dict | list | None:
        """Convert freeform (classic editor) block.

        Classic blocks contain arbitrary HTML that needs to be parsed.
        Returns raw block for now - will be enhanced in S4.
        """
        if not content.strip():
            return None

        # クラシックブロックの中身をパース
        # 再帰的にGutenbergブロックがあればパース
        if "<!-- wp:" in content:
            inner_blocks = []
            for match in self.GUTENBERG_PATTERN.finditer(content):
                block_name = match.group(1)
                attrs_str = match.group(2)
                inner_content = match.group(3)

                try:
                    attrs = json.loads(attrs_str) if attrs_str else {}
                except json.JSONDecodeError:
                    attrs = {}

                block = self._convert_block(block_name, attrs, inner_content, depth + 1)
                if block:
                    if isinstance(block, list):
                        inner_blocks.extend(block)
                    else:
                        inner_blocks.append(block)

            if inner_blocks:
                return inner_blocks

        # Gutenbergブロックがなければ通常HTMLとしてパース（ショートコード対応）
        return self._parse_html(content.strip())

    def _convert_nested(self, content: str, depth: int) -> list | dict | None:
        """Convert nested blocks (columns, group, cover).

        Nested blocks are flattened - inner blocks are extracted.
        """
        if depth > self.MAX_DEPTH:
            return {"type": "raw", "data": {"html": content.strip()}}

        blocks = []

        # 内部のブロックを再帰的にパース
        for match in self.GUTENBERG_PATTERN.finditer(content):
            block_name = match.group(1)
            attrs_str = match.group(2)
            inner_content = match.group(3)

            try:
                attrs = json.loads(attrs_str) if attrs_str else {}
            except json.JSONDecodeError:
                attrs = {}

            block = self._convert_block(block_name, attrs, inner_content, depth + 1)
            if block:
                if isinstance(block, list):
                    blocks.extend(block)
                else:
                    blocks.append(block)

        if blocks:
            return blocks

        # 内部ブロックがなければrawで保持
        if content.strip():
            return {"type": "raw", "data": {"html": content.strip()}}

        return None

    def _process_shortcodes(self, html: str) -> list[dict]:
        """Process WordPress shortcodes in HTML.

        Args:
            html: HTML content that may contain shortcodes

        Returns:
            List of Editor.js blocks from shortcodes
        """
        blocks = []
        last_end = 0

        for match in self.SHORTCODE_PATTERN.finditer(html):
            # ショートコード前のテキスト
            before = html[last_end : match.start()].strip()
            if before:
                # 前のテキストをHTMLとしてパース（再帰防止フラグ付き）
                before_blocks = self._parse_html(before, _from_shortcode=True)
                blocks.extend(before_blocks)

            # ショートコード変換
            shortcode = match.group(1).lower()
            attrs_str = match.group(2) or ""
            content = match.group(3) or ""

            attrs = self._parse_shortcode_attrs(attrs_str)
            block = self._convert_shortcode(shortcode, attrs, content)
            if block:
                if isinstance(block, list):
                    blocks.extend(block)
                else:
                    blocks.append(block)

            last_end = match.end()

        # ショートコード後のテキスト
        after = html[last_end:].strip()
        if after:
            after_blocks = self._parse_html(after, _from_shortcode=True)
            blocks.extend(after_blocks)

        return blocks

    def _parse_shortcode_attrs(self, attrs_str: str) -> dict:
        """Parse shortcode attributes string.

        Args:
            attrs_str: ' attr1="value1" attr2="value2"'

        Returns:
            {"attr1": "value1", "attr2": "value2"}
        """
        attrs = {}
        # attr="value" or attr='value' or attr=value
        pattern = re.compile(r'([\w-]+)=["\']([^"\']*)["\']|([\w-]+)=(\S+)')

        for match in pattern.finditer(attrs_str):
            if match.group(1):
                attrs[match.group(1)] = match.group(2)
            elif match.group(3):
                attrs[match.group(3)] = match.group(4)

        return attrs

    def _convert_shortcode(self, name: str, attrs: dict, content: str) -> dict | list | None:
        """Convert a single shortcode to Editor.js block.

        Args:
            name: Shortcode name (e.g., "gallery", "video")
            attrs: Parsed attributes
            content: Content between [shortcode] and [/shortcode]

        Returns:
            Editor.js block(s) or None
        """
        if name == "gallery":
            return self._convert_gallery_shortcode(attrs)
        elif name == "video":
            return self._convert_video_shortcode(attrs)
        elif name == "audio":
            return self._convert_audio_shortcode(attrs)
        elif name == "embed":
            return self._convert_embed_shortcode(content)
        elif name == "caption":
            return self._convert_caption_shortcode(content, attrs)
        elif name == "playlist":
            return self._convert_playlist_shortcode(attrs)

        # 未対応ショートコード → raw
        shortcode_text = f"[{name}"
        if attrs:
            for k, v in attrs.items():
                shortcode_text += f' {k}="{v}"'
        shortcode_text += "]"
        if content:
            shortcode_text += f"{content}[/{name}]"

        return {"type": "raw", "data": {"html": shortcode_text}}

    def _convert_gallery_shortcode(self, attrs: dict) -> dict | None:
        """Convert [gallery] shortcode.

        [gallery ids="1,2,3" columns="3"]
        """
        ids_str = attrs.get("ids", "")
        if not ids_str:
            return None

        # IDをリストに変換
        ids = [img_id.strip() for img_id in ids_str.split(",") if img_id.strip()]
        if not ids:
            return None

        # columns属性（不正値はデフォルト3）
        try:
            columns = int(attrs.get("columns", 3))
        except (ValueError, TypeError):
            columns = 3

        # Editor.js galleryブロック形式
        # Note: 実際のURLはインポート時にメディアIDから解決する必要がある
        return {
            "type": "gallery",
            "data": {
                "images": [{"id": img_id, "url": f"#media-{img_id}"} for img_id in ids],
                "columns": columns,
            },
        }

    def _convert_video_shortcode(self, attrs: dict) -> dict | None:
        """Convert [video] shortcode.

        [video src="url" poster="url"]
        """
        src = attrs.get("src", "")
        if not src:
            return None

        return {
            "type": "embed",
            "data": {
                "service": "video",
                "source": src,
                "embed": src,
                "caption": "",
            },
        }

    def _convert_audio_shortcode(self, attrs: dict) -> dict | None:
        """Convert [audio] shortcode.

        [audio src="url"]
        """
        src = attrs.get("src", "")
        if not src:
            return None

        return {
            "type": "embed",
            "data": {
                "service": "audio",
                "source": src,
                "embed": src,
                "caption": "",
            },
        }

    def _convert_embed_shortcode(self, content: str) -> dict | None:
        """Convert [embed] shortcode.

        [embed]https://youtube.com/watch?v=xxx[/embed]
        """
        url = content.strip()
        if not url:
            return None

        # サービス判定
        service = ""
        embed_url = url
        if "youtube" in url or "youtu.be" in url:
            service = "youtube"
            if "watch?v=" in url:
                video_id = url.split("watch?v=")[-1].split("&")[0]
                embed_url = f"https://www.youtube.com/embed/{video_id}"
            elif "youtu.be/" in url:
                video_id = url.split("youtu.be/")[-1].split("?")[0]
                embed_url = f"https://www.youtube.com/embed/{video_id}"
        elif "vimeo" in url:
            service = "vimeo"
        elif "twitter" in url or "x.com" in url:
            service = "twitter"
        elif "spotify" in url:
            service = "spotify"

        return {
            "type": "embed",
            "data": {
                "service": service,
                "source": url,
                "embed": embed_url,
                "caption": "",
            },
        }

    def _convert_caption_shortcode(self, content: str, attrs: dict) -> dict | None:
        """Convert [caption] shortcode.

        [caption]<img src="...">Caption text[/caption]
        """
        if not content.strip():
            return None

        soup = BeautifulSoup(content, "html.parser")
        img = soup.find("img")

        if img:
            url = img.get("src", "")
            alt = img.get("alt", "")
            # img以外のテキストがキャプション
            img.extract()
            caption = soup.get_text(strip=True) or alt

            if url:
                return {
                    "type": "image",
                    "data": {"file": {"url": url}, "caption": caption},
                }

        # imgがなければraw
        return {"type": "raw", "data": {"html": content.strip()}}

    def _convert_playlist_shortcode(self, attrs: dict) -> dict | None:
        """Convert [playlist] shortcode.

        [playlist ids="1,2,3" type="audio"]
        """
        ids_str = attrs.get("ids", "")
        if not ids_str:
            return None

        playlist_type = attrs.get("type", "audio")

        # プレイリストはrawで保持（Editor.jsに対応ブロックなし）
        return {
            "type": "raw",
            "data": {
                "html": f'<div class="wp-playlist" data-type="{playlist_type}" data-ids="{ids_str}"></div>'
            },
        }

    def _parse_html(self, html: str, _from_shortcode: bool = False) -> list[dict]:
        """Parse classic HTML (no Gutenberg).

        Args:
            html: Plain HTML without Gutenberg comments
            _from_shortcode: Internal flag to prevent infinite recursion

        Returns:
            List of Editor.js blocks
        """
        # ショートコード検出（再帰呼び出し時はスキップ）
        if not _from_shortcode and "[" in html and "]" in html:
            if self.SHORTCODE_PATTERN.search(html):
                return self._process_shortcodes(html)

        soup = BeautifulSoup(html, "html.parser")
        blocks = []

        # トップレベル要素を順に処理
        for element in soup.children:
            if isinstance(element, str):
                # テキストノード
                text = element.strip()
                if text:
                    blocks.append({"type": "paragraph", "data": {"text": text}})
                continue

            if element.name is None:
                continue

            block = self._html_element_to_block(element)
            if block:
                if isinstance(block, list):
                    blocks.extend(block)
                else:
                    blocks.append(block)

        # ブロックが生成されなかった場合はrawで保持
        if not blocks and html.strip():
            blocks = [{"type": "raw", "data": {"html": html.strip()}}]

        return blocks

    def _html_element_to_block(self, element) -> dict | list | None:
        """Convert a single HTML element to Editor.js block."""
        tag = element.name

        if tag == "p":
            text = str(element.decode_contents())
            if text.strip():
                return {"type": "paragraph", "data": {"text": text}}

        elif tag in ("h1", "h2", "h3", "h4", "h5", "h6"):
            level = int(tag[1])
            text = element.get_text(strip=True)
            if text:
                return {"type": "header", "data": {"text": text, "level": level}}

        elif tag == "ul":
            items = []
            for li in element.find_all("li", recursive=False):
                items.append(str(li.decode_contents()))
            if items:
                return {"type": "list", "data": {"style": "unordered", "items": items}}

        elif tag == "ol":
            items = []
            for li in element.find_all("li", recursive=False):
                items.append(str(li.decode_contents()))
            if items:
                return {"type": "list", "data": {"style": "ordered", "items": items}}

        elif tag == "img":
            url = element.get("src", "")
            alt = element.get("alt", "")
            if url:
                return {
                    "type": "image",
                    "data": {"file": {"url": url}, "caption": alt},
                }

        elif tag == "figure":
            img = element.find("img")
            if img:
                url = img.get("src", "")
                alt = img.get("alt", "")
                figcaption = element.find("figcaption")
                caption = figcaption.get_text(strip=True) if figcaption else alt
                if url:
                    return {
                        "type": "image",
                        "data": {"file": {"url": url}, "caption": caption},
                    }

        elif tag == "blockquote":
            p = element.find("p")
            text = str(p.decode_contents()) if p else element.get_text(strip=True)
            cite = element.find("cite")
            caption = cite.get_text(strip=True) if cite else ""
            if text:
                return {"type": "quote", "data": {"text": text, "caption": caption}}

        elif tag == "pre":
            code = element.find("code")
            code_text = code.get_text() if code else element.get_text()
            if code_text:
                return {"type": "code", "data": {"code": code_text}}

        elif tag == "hr":
            return {"type": "delimiter", "data": {}}

        elif tag == "table":
            rows = []
            for tr in element.find_all("tr"):
                cells = []
                for cell in tr.find_all(["td", "th"]):
                    cells.append(str(cell.decode_contents()))
                if cells:
                    rows.append(cells)
            if rows:
                return {"type": "table", "data": {"content": rows}}

        elif tag == "iframe":
            src = element.get("src", "")
            if src:
                service = ""
                if "youtube" in src:
                    service = "youtube"
                elif "vimeo" in src:
                    service = "vimeo"
                return {
                    "type": "embed",
                    "data": {"service": service, "source": src, "embed": src, "caption": ""},
                }

        elif tag == "div":
            # divの中身を再帰的にパース
            inner_blocks = []
            for child in element.children:
                if isinstance(child, str):
                    text = child.strip()
                    if text:
                        inner_blocks.append({"type": "paragraph", "data": {"text": text}})
                elif child.name:
                    block = self._html_element_to_block(child)
                    if block:
                        if isinstance(block, list):
                            inner_blocks.extend(block)
                        else:
                            inner_blocks.append(block)
            if inner_blocks:
                return inner_blocks
            # 中身がなければrawで保持
            content = str(element.decode_contents()).strip()
            if content:
                return {"type": "raw", "data": {"html": str(element)}}

        # 未対応タグはrawで保持
        content = str(element).strip()
        if content:
            return {"type": "raw", "data": {"html": content}}

        return None


# シングルトン
block_converter = BlockConverter()
